from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest
import rich.console
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_frame_equal
from xorq.vendor import ibis


if TYPE_CHECKING:
    pass


@pytest.fixture
def new_schema():
    return xo.schema([("a", "string"), ("b", "bool"), ("c", "int32")])


def _create_temp_table_with_schema(con, temp_table_name, schema, data=None):
    temporary = con.create_table(temp_table_name, schema=schema)
    assert temporary.to_pandas().empty

    if data is not None and isinstance(data, pd.DataFrame):
        assert not data.empty
        tmp = con.create_table(temp_table_name, data, overwrite=True)
        result = tmp.to_pandas()
        assert len(result) == len(data.index)
        tm.assert_frame_equal(
            result.sort_values(result.columns[0]).reset_index(drop=True),
            data.sort_values(result.columns[0]).reset_index(drop=True),
        )
        return tmp

    return temporary


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(
            xo.memtable([(1, 2.0, "3")], columns=list("abc")),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")),
            id="simple",
        ),
        param(
            xo.memtable([(1, 2.0, "3")]),
            pd.DataFrame([(1, 2.0, "3")], columns=["col0", "col1", "col2"]),
            id="simple_auto_named",
        ),
        param(
            xo.memtable(
                [(1, 2.0, "3")],
                schema=xo.schema({"a": "int8", "b": "float32", "c": "string"}),
            ),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")).astype(
                {"a": "int8", "b": "float32"}
            ),
            id="simple_schema",
        ),
        param(
            xo.memtable(
                pd.DataFrame({"a": [1], "b": [2.0], "c": ["3"]}).astype(
                    {"a": "int8", "b": "float32"}
                )
            ),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")).astype(
                {"a": "int8", "b": "float32"}
            ),
            id="dataframe",
        ),
        param(
            xo.memtable([{"a": 1}, {"a": 2}]),
            pd.DataFrame({"a": [1, 2]}),
            id="list_of_dicts",
        ),
    ],
)
def test_in_memory_table(con, expr, expected):
    result = con.execute(expr)
    assert_frame_equal(result, expected)


def test_filter_memory_table(con):
    t = xo.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.filter(t.x > 1)
    expected = pd.DataFrame({"x": [3, 5], "y": [4, 6]})
    result = con.execute(expr)
    assert_frame_equal(result, expected)


def test_agg_memory_table(con):
    t = xo.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.x.count()
    result = con.execute(expr)
    assert result == 3


def test_self_join_memory_table(con):
    t = xo.memtable({"x": [1, 2], "y": [2, 1], "z": ["a", "b"]})
    t_view = t.view()
    expr = t.join(t_view, t.x == t_view.y).select("x", "y", "z", "z_right")
    result = con.execute(expr).sort_values("x").reset_index(drop=True)
    expected = pd.DataFrame(
        {"x": [1, 2], "y": [2, 1], "z": ["a", "b"], "z_right": ["b", "a"]}
    )
    assert_frame_equal(result, expected)


@pytest.mark.parametrize("dtype", [None, "f8"])
def test_dunder_array_table(alltypes, dtype):
    expr = alltypes.group_by("string_col").int_col.sum().order_by("string_col")
    result = np.asarray(expr, dtype=dtype)
    expected = np.asarray(expr.execute(), dtype=dtype)
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("dtype", [None, "f8"])
def test_dunder_array_column(alltypes, dtype):
    from xorq.vendor.ibis import _

    expr = alltypes.group_by("string_col").agg(int_col=_.int_col.sum()).int_col
    result = np.sort(np.asarray(expr, dtype=dtype))
    expected = np.sort(np.asarray(xo.execute(expr), dtype=dtype))
    np.testing.assert_array_equal(result, expected)


@pytest.mark.xfail
@pytest.mark.parametrize("interactive", [True, False])
def test_repr(alltypes, interactive, monkeypatch):
    monkeypatch.setattr(xo.options, "interactive", interactive)

    expr = alltypes.select("date_string_col")

    s = repr(expr)
    # no control characters
    assert all(c.isprintable() or c in "\n\r\t" for c in s)
    if interactive:
        assert "/" in s
    else:
        assert "/" not in s


@pytest.mark.xfail
@pytest.mark.parametrize("show_types", [True, False])
def test_interactive_repr_show_types(alltypes, show_types, monkeypatch):
    monkeypatch.setattr(xo.options, "interactive", True)
    monkeypatch.setattr(xo.options.repr.interactive, "show_types", show_types)

    expr = alltypes.select("id")
    s = repr(expr)
    if show_types:
        assert "int" in s
    else:
        assert "int" not in s


@pytest.mark.xfail
@pytest.mark.parametrize("is_jupyter", [True, False])
def test_interactive_repr_max_columns(alltypes, is_jupyter, monkeypatch):
    monkeypatch.setattr(xo.options, "interactive", True)

    cols = {f"c_{i}": ibis._.id + i for i in range(50)}
    expr = alltypes.mutate(**cols).select(*cols)

    console = rich.console.Console(force_jupyter=is_jupyter, width=80)
    options = console.options.copy()

    # max_columns = 0
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_0 " in text
    if is_jupyter:
        # All columns are written
        assert " c_49 " in text
    else:
        # width calculations truncate well before 20 columns
        assert " c_19 " not in text

    # max_columns = 3
    monkeypatch.setattr(xo.options.repr.interactive, "max_columns", 3)
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_2 " in text
    assert " c_3 " not in text

    # max_columns = None
    monkeypatch.setattr(xo.options.repr.interactive, "max_columns", None)
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_0 " in text
    if is_jupyter:
        # All columns written
        assert " c_49 " in text
    else:
        # width calculations still truncate
        assert " c_19 " not in text


@pytest.mark.xfail
@pytest.mark.parametrize("expr_type", ["table", "column"])
@pytest.mark.parametrize("interactive", [True, False])
def test_repr_mimebundle(alltypes, interactive, expr_type, monkeypatch):
    monkeypatch.setattr(xo.options, "interactive", interactive)

    if expr_type == "column":
        expr = alltypes.date_string_col
    else:
        expr = alltypes.select("date_string_col")

    reprs = expr._repr_mimebundle_(include=["text/plain", "text/html"], exclude=[])
    for fmt in ["text/plain", "text/html"]:
        if interactive:
            assert "r0.date_string_col" not in reprs[fmt]
        else:
            assert "r0.date_string_col" in reprs[fmt]


@pytest.mark.parametrize(
    "option", ["max_rows", "max_length", "max_string", "max_depth"]
)
def test_ibis_config_wrapper(option, monkeypatch):
    from xorq.vendor import ibis

    xorq_option_value = getattr(xo.options.repr.interactive, option)
    assert xorq_option_value == getattr(ibis.options.repr.interactive, option)

    monkeypatch.setattr(xo.options.repr.interactive, option, xorq_option_value + 1)
    assert getattr(ibis.options.repr.interactive, option) == xorq_option_value + 1



from __future__ import annotations

import ast

import toolz


class NameFinder:
    """Helper class to find the unique names in an AST."""

    __slots__ = ()

    def find(self, node):
        typename = type(node).__name__
        method = getattr(self, f"find_{typename}", None)
        if method is None:
            fields = getattr(node, "_fields", None)
            if fields is None:
                return
            for field in fields:
                value = getattr(node, field)
                yield from self.find(value)
        else:
            yield from method(node)

    def find_Name(self, node):
        # TODO not sure if this is robust to scope changes
        yield node

    def find_list(self, node):
        return list(toolz.concat(map(self.find, node)))

    def find_Call(self, node):
        if not isinstance(node.func, ast.Name):
            fields = node._fields
        else:
            fields = [field for field in node._fields if field != "func"]
        return toolz.concat(map(self.find, (getattr(node, field) for field in fields)))


def find_names(node: ast.AST) -> list[ast.Name]:
    """Return the unique `ast.Name` instances in an AST.

    Examples
    --------
    >>> import ast
    >>> node = ast.parse("a + b")
    >>> names = find_names(node)
    >>> names
    [<....Name object at 0x...>, <....Name object at 0x...>]
    >>> names[0].id
    'a'
    >>> names[1].id
    'b'

    """
    return list(
        toolz.unique(
            filter(None, NameFinder().find(node)),
            key=lambda node: (node.id, type(node.ctx)),
        )
    )



import re

import xorq.api as xo
from xorq.api import _
from xorq.common.utils.defer_utils import deferred_read_csv


def test_read(csv_dir):
    csv_name = "astronauts"
    csv_path = csv_dir / f"{csv_name}.csv"

    # we can work with a pandas expr without having read it yet
    pd_con = xo.pandas.connect()

    expr = deferred_read_csv(con=pd_con, path=csv_path, table_name=csv_name).filter(
        _.number > 6
    )

    pattern = r"Read\[name=astronauts, method_name=read_csv, source=pandas-\d+\]"

    assert re.search(pattern, repr(expr))



from __future__ import annotations

import contextlib
import functools
from abc import (
    abstractmethod,
)

import dask
from attr import (
    frozen,
)

import xorq.common.utils.dask_normalize  # noqa: F401
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.dask_normalize.dask_normalize_expr import (
    normalize_backend,
    normalize_read,
    normalize_remote_table,
)
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    patch_normalize_op_caching,
    patch_normalize_token,
)
from xorq.expr.relations import (
    Read,
    RemoteTable,
)
from xorq.vendor.ibis.expr import types as ir


@frozen
class CacheStrategy:
    @abstractmethod
    def calc_key(self, expr):
        pass

    def __dask_tokenize__(self):
        return (type(self).__name__,)


@frozen
class ModificationTimeStrategy(CacheStrategy):
    def calc_key(self, expr: ir.Expr):
        return expr.ls.tokenized


@frozen
class SnapshotStrategy(CacheStrategy):
    def calc_key(self, expr: ir.Expr):
        # can we cache this?
        with self.normalization_context(expr):
            replaced = self.replace_remote_table(expr)
            tokenized = replaced.ls.tokenized
            return "-".join(("snapshot", tokenized))

    @contextlib.contextmanager
    def normalization_context(self, expr):
        # patch_normalize_op_caching memoizes normalize_op by (op, compiler).
        # Without it, tokenizing depth-n pipeline expressions is O(2^n) because
        # normalize_remote_table recursively tokenizes remote_expr and
        # normalize_scalar_udf recursively tokenizes computed_kwargs_expr, both
        # of which share sub-expressions that get re-tokenized without caching.
        typs = map(type, expr.ls.backends)
        with patch_normalize_op_caching():
            with patch_normalize_token(*typs, f=self.normalize_backend):
                with patch_normalize_token(
                    ops.DatabaseTable,
                    f=self.normalize_databasetable,
                ):
                    with patch_normalize_token(
                        Read,
                        f=self.cached_normalize_read,
                    ):
                        yield

    @staticmethod
    @functools.cache
    def cached_normalize_read(op):
        return normalize_read(op)

    @staticmethod
    @functools.cache
    def cached_replace_remote_table(op):
        def rename_remote_table(node, _, **kwargs):
            if isinstance(node, RemoteTable):
                # FIXME: how to verify that we're always within a self.normalization_context?
                name = dask.base.tokenize(node)
                rt = RemoteTable(
                    name=name,
                    schema=node.schema,
                    source=node.source,
                    remote_expr=node.remote_expr,
                    namespace=node.namespace,
                )
                return rt
            else:
                return node.__recreate__(kwargs)

        return op.replace(rename_remote_table)

    def replace_remote_table(self, expr):
        """replace remote table with deterministic name ***strictly for key calculation***"""
        if expr.op().find(RemoteTable):
            expr = self.cached_replace_remote_table(expr.op()).to_expr()
        return expr

    @staticmethod
    def normalize_backend(con):
        name = con.name
        if name in ("pandas", "duckdb", "datafusion", "let"):
            return (name, None)
        else:
            return normalize_backend(con)

    @staticmethod
    def normalize_databasetable(dt):
        if isinstance(dt, RemoteTable):
            # one alternative is to explicitly iterate over the fields name, schema, source, namespace
            # but explicit is better than implicit, additionally the name is not a safe bet for caching
            # RemoteTable
            return normalize_remote_table(dt)
        else:
            keys = ["name", "schema", "source", "namespace"]
            return dask.tokenize._normalize_seq_func(
                (key, getattr(dt, key)) for key in keys
            )



import xorq
import xorq.api as xo
from xorq.backends.xorq.datafusion import Backend


def test_version():
    assert xorq.__version__ == Backend().version


def test_context_name():
    con = xo.connect()
    assert "let.SessionContext" in str(type(con.con))



import os
import sys

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as OTLPSpanExporterGRPC,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
)

from xorq.common.utils.env_utils import (
    EnvConfigable,
    env_templates_dir,
)


def localhost_and_listening(uri):
    import socket
    import urllib

    parsed = urllib.parse.urlparse(uri)
    localhost = "localhost"
    if parsed.hostname == localhost:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind((localhost, parsed.port))
        except OSError:
            return True
        else:
            return False
    return None


OTELConfig = EnvConfigable.subclass_from_env_file(
    env_templates_dir.joinpath(".env.otel.template")
)
otel_config = OTELConfig.from_env()


def get_otlp_exporter():
    """Create OTLP exporter based on protocol configuration.

    SDK auto-configures from standard OTEL environment variables:
    - OTEL_EXPORTER_OTLP_TRACES_ENDPOINT or OTEL_EXPORTER_OTLP_ENDPOINT
    - OTEL_EXPORTER_OTLP_PROTOCOL (grpc or http/protobuf)
    - OTEL_EXPORTER_OTLP_HEADERS
    - OTEL_EXPORTER_OTLP_TIMEOUT
    """
    protocol = os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "http/protobuf")

    if protocol == "grpc":
        return OTLPSpanExporterGRPC()
    else:
        return OTLPSpanExporter()


resource_attributes = {
    SERVICE_NAME: otel_config.OTEL_SERVICE_NAME,
    **({"execution.id": eid} if (eid := otel_config.OTEL_EXECUTION_ID) else {}),
}

resource = Resource(resource_attributes)

provider = TracerProvider(resource=resource)

traces_endpoint = otel_config.OTEL_EXPORTER_OTLP_TRACES_ENDPOINT

if traces_endpoint and localhost_and_listening(traces_endpoint):
    processor = BatchSpanProcessor(get_otlp_exporter())
else:
    processor = BatchSpanProcessor(
        ConsoleSpanExporter(
            out=sys.stdout
            if otel_config.get("OTEL_EXPORTER_CONSOLE_FALLBACK")
            else open(os.devnull, "w")
        )
    )
provider.add_span_processor(processor)
trace.set_tracer_provider(provider)

tracer = trace.get_tracer("xorq.tracer")



"""Utilities for generating self-signed TLS certificates using pyca/cryptography."""

import datetime
import operator
import socket

from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
    optional,
)
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import (
    padding,
    rsa,
)
from cryptography.hazmat.primitives.serialization import (
    BestAvailableEncryption,
    Encoding,
    NoEncryption,
    PrivateFormat,
    load_pem_private_key,
)
from cryptography.x509.oid import NameOID


try:
    utc = datetime.UTC
except AttributeError:
    utc = datetime.timezone.utc

pem_encode_cert = operator.methodcaller("public_bytes", Encoding.PEM)
pem_decode_cert = x509.load_pem_x509_certificate


def pem_encode_private_key(private_key, password=None):
    encryption = (
        NoEncryption() if password is None else BestAvailableEncryption(password)
    )
    return private_key.private_bytes(
        Encoding.PEM, PrivateFormat.TraditionalOpenSSL, encryption
    )


def pem_decode_private_key(byts, password=None):
    return load_pem_private_key(byts, password)


def make_tls_cert_and_private_key(
    common_name, *sans, duration=datetime.timedelta(5 * 365)
) -> tuple[x509.Certificate, rsa.RSAPrivateKey]:
    # Generate RSA private key (2048-bit)
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]
    )

    # Build certificate with extensions
    now = datetime.datetime.now(utc)
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(now)
        .not_valid_after(now + duration)
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(san) for san in sans]),
            critical=False,
        )
        .add_extension(
            # can sign leaf but not intermediary
            x509.BasicConstraints(ca=True, path_length=0),
            critical=True,
        )
        .sign(
            private_key,
            hashes.SHA256(),
            default_backend(),
        )
    )

    return (cert, private_key)


@frozen
class TLSCert:
    cert = field(validator=instance_of(x509.Certificate))
    private_key = field(
        validator=instance_of(rsa.RSAPrivateKey),
        eq=operator.methodcaller("private_numbers"),
    )

    def __equals__(self, other):
        return (
            self.cert == other.cert
            and self.private_key.private_numbers()
            == other.private_key.private_numbers()
        )

    @property
    def cert_bytes(self):
        return pem_encode_cert(self.cert)

    @property
    def private_key_bytes(self):
        return pem_encode_private_key(self.private_key)

    @property
    def tls_certificates(self):
        return [(self.cert_bytes, self.private_key_bytes)]

    def sign_csr(self, csr):
        return sign_csr(self.cert, self.private_key, csr)

    def create_csr(self):
        return create_csr(self.cert, self.private_key)

    def verify(self, other):
        verify(self.cert, other.cert)

    def signed_with(self, ca_tlscert):
        csr = self.create_csr()
        cert = ca_tlscert.sign_csr(csr)
        return type(self)(cert, self.private_key)

    def to_disk(self, cert_path, private_key_path, password=None):
        cert_path.write_bytes(pem_encode_cert(self.cert))
        private_key_path.write_bytes(pem_encode_private_key(self.private_key, password))

    @classmethod
    def from_disk(cls, cert_path, private_key_path, password=None):
        return cls(
            pem_decode_cert(cert_path.read_bytes()),
            pem_decode_private_key(private_key_path.read_bytes(), password),
        )

    @classmethod
    def from_common_name(
        cls,
        common_name=socket.gethostname(),  # noqa: B008
        *,
        sans=(),
        sign_with=None,
    ):
        (cert, private_key) = make_tls_cert_and_private_key(common_name, *sans)
        self = cls(cert, private_key)
        if sign_with:
            self = self.signed_with(sign_with)
        return self


@frozen
class TLSKwargs:
    verify_client = field(validator=instance_of(bool))
    ca_tlscert = field(validator=instance_of(TLSCert))
    server_tlscert = field(validator=instance_of(TLSCert))
    client_tlscert = field(
        validator=optional(instance_of(TLSCert)),
        default=None,
    )

    def __attrs_post_init__(self):
        self.ca_tlscert.verify(self.server_tlscert)
        if self.verify_client:
            assert self.client_tlscert is not None
            self.ca_tlscert.verify(self.client_tlscert)

    @property
    def server_kwargs(self):
        return {
            "tls_certificates": self.server_tlscert.tls_certificates,
            "root_certificates": self.ca_tlscert.cert_bytes,
        }

    @property
    def client_kwargs(self):
        client_kwargs = {
            "tls_root_certs": self.ca_tlscert.cert_bytes,
        }

        if self.verify_client:
            client_kwargs |= {
                "cert_chain": self.client_tlscert.cert_bytes,
                "private_key": self.client_tlscert.private_key_bytes,
            }

        return client_kwargs

    @classmethod
    def from_kwargs(cls, ca_kwargs, server_kwargs, client_kwargs, verify_client=True):
        if server_kwargs.get(
            "common_name"
        ) != "localhost" or "localhost" not in server_kwargs.get("sans", ()):
            # FIXME: issue a warning
            pass
        ca_tlscert = TLSCert.from_common_name(**ca_kwargs)
        server_tlscert = TLSCert.from_common_name(sign_with=ca_tlscert, **server_kwargs)
        client_tlscert = TLSCert.from_common_name(sign_with=ca_tlscert, **client_kwargs)

        return cls(verify_client, ca_tlscert, server_tlscert, client_tlscert)

    @classmethod
    def from_common_name(cls, verify_client=True, common_name=socket.gethostname()):  # noqa: B008
        ca_kwargs = {
            "common_name": "root_cert",
        }
        server_kwargs = {
            "common_name": common_name,
            "sans": ("localhost",),
        }
        client_kwargs = {
            "common_name": "client",
        }

        return cls.from_kwargs(
            ca_kwargs, server_kwargs, client_kwargs, verify_client=verify_client
        )

    @classmethod
    def from_disk(
        cls,
        ca_cert_path,
        ca_private_key_path,
        server_cert_path,
        server_private_key_path,
        client_cert_path=None,
        client_private_key_path=None,
        verify_client=True,
        ca_password=None,
        server_password=None,
        client_password=None,
    ):
        """
        Create TLSKwargs by loading certificates and private keys from disk.

        Args:
            ca_cert_path: Path to CA certificate file
            ca_private_key_path: Path to CA private key file
            server_cert_path: Path to server certificate file
            server_private_key_path: Path to server private key file
            client_cert_path: Path to client certificate file (optional)
            client_private_key_path: Path to client private key file (optional)
            verify_client: Whether client verification is required
            ca_password: Password for CA private key (optional)
            server_password: Password for server private key (optional)
            client_password: Password for client private key (optional)

        Returns:
            TLSKwargs instance with certificates loaded from disk
        """
        # Load CA certificate
        ca_tlscert = TLSCert.from_disk(ca_cert_path, ca_private_key_path, ca_password)

        # Load server certificate
        server_tlscert = TLSCert.from_disk(
            server_cert_path, server_private_key_path, server_password
        )

        # Load client certificate if paths provided or if client verification is required
        client_tlscert = None
        if client_cert_path is not None and client_private_key_path is not None:
            client_tlscert = TLSCert.from_disk(
                client_cert_path, client_private_key_path, client_password
            )
        elif verify_client:
            raise ValueError(
                "Client certificate paths must be provided when verify_client=True"
            )

        return cls(
            verify_client=verify_client,
            ca_tlscert=ca_tlscert,
            server_tlscert=server_tlscert,
            client_tlscert=client_tlscert,
        )

    def to_disk(
        self,
        ca_cert_path,
        ca_private_key_path,
        server_cert_path,
        server_private_key_path,
        client_cert_path=None,
        client_private_key_path=None,
        ca_password=None,
        server_password=None,
        client_password=None,
    ):
        """
        Save all TLS certificates and private keys to disk.

        Args:
            ca_cert_path: Path where CA certificate will be saved
            ca_private_key_path: Path where CA private key will be saved
            server_cert_path: Path where server certificate will be saved
            server_private_key_path: Path where server private key will be saved
            client_cert_path: Path where client certificate will be saved (optional)
            client_private_key_path: Path where client private key will be saved (optional)
            ca_password: Password to encrypt CA private key (optional)
            server_password: Password to encrypt server private key (optional)
            client_password: Password to encrypt client private key (optional)

        Raises:
            ValueError: If client paths are not provided when client_tlscert exists
        """
        # Save CA certificate and key
        self.ca_tlscert.to_disk(ca_cert_path, ca_private_key_path, ca_password)

        # Save server certificate and key
        self.server_tlscert.to_disk(
            server_cert_path, server_private_key_path, server_password
        )

        # Save client certificate and key if it exists
        if self.client_tlscert is not None:
            if client_cert_path is None or client_private_key_path is None:
                raise ValueError(
                    "Client certificate paths must be provided when client_tlscert exists"
                )
            self.client_tlscert.to_disk(
                client_cert_path, client_private_key_path, client_password
            )


def get_san(cert):
    cls = x509.SubjectAlternativeName
    san = cls([])
    try:
        ext = cert.extensions.get_extension_for_class(cls)
        san = ext.value
    except Exception:
        pass
    return san


def get_common_name(cert):
    (attr, *rest) = cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)
    assert not rest
    return attr.value


def create_csr(cert, private_key):
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(cert.subject)
        .add_extension(get_san(cert), critical=False)
        .sign(private_key, hashes.SHA256())
    )
    return csr


def sign_csr(
    signing_cert, signing_private_key, csr, duration=datetime.timedelta(days=5 * 365)
):
    now = datetime.datetime.now(utc)
    cert = (
        x509.CertificateBuilder()
        .subject_name(csr.subject)
        .issuer_name(signing_cert.subject)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(now)
        .not_valid_after(now + duration)
        .add_extension(get_san(csr), critical=False)
        .sign(signing_private_key, hashes.SHA256())
    )
    return cert


def verify(ca_cert, other_cert):
    public_key = ca_cert.public_key()
    public_key.verify(
        other_cert.signature,
        other_cert.tbs_certificate_bytes,
        padding.PKCS1v15(),
        other_cert.signature_hash_algorithm,
    )



from __future__ import annotations

import ast
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Callable


def matches(value: ast.AST, pattern: ast.AST) -> bool:
    """Check whether `value` matches `pattern`."""
    # types must match exactly
    if type(value) is not type(pattern):
        return False

    # primitive value, such as None, True, False etc
    if not isinstance(value, ast.AST) and not isinstance(pattern, ast.AST):
        return value == pattern

    fields = [
        (field, getattr(pattern, field))
        for field in pattern._fields
        if hasattr(pattern, field)
    ]
    return all(
        matches(getattr(value, field_name), field_value)
        for field_name, field_value in fields
    )


class Rewriter:
    """AST pattern matcher to enable rewrite rules."""

    def __init__(self):
        self.funcs: list[tuple[ast.AST, Callable[[ast.expr], ast.expr]]] = []

    def register(self, pattern):
        def wrapper(f):
            self.funcs.append((pattern, f))
            return f

        return wrapper

    def __call__(self, node):
        # TODO: more efficient way of doing this?
        for pattern, func in self.funcs:
            if matches(node, pattern):
                return func(node)
        return node


rewrite = Rewriter()



import warnings
from typing import Any, Dict, List, Tuple, TypedDict

import toolz

import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.common.exceptions import XorqError
from xorq.common.utils.graph_utils import walk_nodes
from xorq.expr.relations import Read, RemoteTable


class QueryInfo(TypedDict):
    engine: str
    profile_name: str
    sql: str


class SQLPlans(TypedDict):
    queries: Dict[str, QueryInfo]


class DeferredReadsPlan(TypedDict):
    reads: Dict[str, QueryInfo]


def to_sql(expr: ir.Expr) -> str:
    try:
        compiler_provider = expr.ls.uncached._find_backend(use_default=True)
        if getattr(compiler_provider, "compiler", None) is None:
            warnings.warn(
                f"{compiler_provider} is not a SQL backend, so no SQL string will be generated",
                stacklevel=2,
            )
            return ""
    except XorqError:
        pass

    return ibis.to_sql(expr.ls.uncached)


def find_relations(expr: ir.Expr) -> List[str]:
    def get_name(node):
        name = None
        if isinstance(node, RemoteTable):
            name = node.name
        elif isinstance(node, Read):
            name = node.make_unbound_dt().name
        elif isinstance(node, ops.DatabaseTable):
            name = node.name
        return name

    node_types = (RemoteTable, Read, ops.DatabaseTable)
    nodes = walk_nodes(node_types, expr)
    relations = sorted(set(filter(None, map(get_name, nodes))))
    return relations


def find_tables(expr: ir.Expr) -> Tuple[Dict[str, QueryInfo], Dict[str, QueryInfo]]:
    def get_remote_table_backend(node):
        return node.remote_expr._find_backend()

    grouped = toolz.groupby(type, walk_nodes((RemoteTable, Read), expr))
    remote_tables: Dict[str, QueryInfo] = {
        node.name: {
            "engine": backend.name,
            "profile_name": backend._profile.hash_name,
            "relations": find_relations(node.remote_expr),
            "sql": to_sql(node.remote_expr).strip(),
            "options": {},
        }
        for node in grouped.get(RemoteTable, ())
        if (backend := get_remote_table_backend(node))
    }
    deferred_reads: Dict[str, QueryInfo] = {
        dt.name: {
            "engine": backend.name,
            "profile_name": backend._profile.hash_name,
            "relations": [dt.name],
            "sql": to_sql(dt.to_expr()).strip(),
            "options": get_read_options(node),
        }
        for node in grouped.get(Read, ())
        if (backend := node.source) and (dt := node.make_unbound_dt())
    }
    remote_tables = dict(sorted(remote_tables.items()))
    deferred_reads = dict(sorted(deferred_reads.items()))
    return remote_tables, deferred_reads


def get_read_options(read_instance) -> Dict[str, Any]:
    read_kwargs_list = [{k: v} for k, v in read_instance.read_kwargs]
    return {
        "method_name": read_instance.method_name,
        "name": read_instance.name,
        "read_kwargs": read_kwargs_list,
    }


def generate_sql_plans(expr: ir.Expr) -> Tuple[SQLPlans, DeferredReadsPlan]:
    remote_tables, deferred_reads = find_tables(expr)
    backend = expr._find_backend()

    queries: Dict[str, QueryInfo] = {
        "main": {
            "engine": backend.name,
            "profile_name": backend._profile.hash_name,
            "relations": find_relations(expr),
            "sql": to_sql(expr).strip(),
            "options": {},
        }
    } | remote_tables

    sql_plans: SQLPlans = {"queries": queries}
    deferred_reads_plans: DeferredReadsPlan = {"reads": deferred_reads}
    return sql_plans, deferred_reads_plans



from pathlib import Path

import pytest

from xorq.common.utils.io_utils import extract_suffix


@pytest.mark.parametrize(
    "input_path,expected",
    [
        # Local and HTTP cases
        ("file.txt", ".txt"),
        (Path("file.txt"), ".txt"),
        ("archive.tar.gz", ".gz"),
        (Path("archive.tar.gz"), ".gz"),
        ("README", ""),
        (Path("README"), ""),
        (".gitignore", ""),
        ("https://example.com/file/data.csv", ".csv"),
        ("ftp://host.org/archive.tar.gz", ".gz"),
        ("https://example.com/file/README", ""),
        ("folder/", ""),
        (Path("folder/"), ""),
        # S3, GS, and GCS cases
        ("s3://my-bucket/data/file.json", ".json"),
        ("s3://bucket/folder/archive.tar.gz", ".gz"),
        ("s3://bucket/folder/", ""),
        ("gs://my-bucket/data/file.csv", ".csv"),
        ("gs://bucket/folder/", ""),
        ("gcs://bucket/file.txt", ".txt"),
        ("gcs://bucket/folder/", ""),
    ],
)
def test_extract_suffix(input_path, expected):
    assert extract_suffix(input_path) == expected



"""Operations for numeric expressions."""

from __future__ import annotations

import operator
from typing import Optional

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.expr.operations.core import Binary, Unary, Value


Integer = Value[dt.Integer]
SoftNumeric = Value[dt.Numeric | dt.Boolean]
StrictNumeric = Value[dt.Numeric]


@public
class NumericBinary(Binary):
    left: SoftNumeric
    right: SoftNumeric


@public
class Add(NumericBinary):
    """Add two values."""

    dtype = rlz.numeric_like("args", operator.add)


@public
class Multiply(NumericBinary):
    """Multiply two values."""

    dtype = rlz.numeric_like("args", operator.mul)


@public
class Power(NumericBinary):
    """Raise the left value to the power of the right value."""

    @property
    def dtype(self):
        dtypes = (arg.dtype for arg in self.args)
        if util.all_of(dtypes, dt.Integer):
            return dt.float64
        else:
            return rlz.highest_precedence_dtype(self.args)


@public
class Subtract(NumericBinary):
    """Subtract the right value from the left value."""

    dtype = rlz.numeric_like("args", operator.sub)


@public
class Divide(NumericBinary):
    """Divide the left value by the right value."""

    dtype = dt.float64


@public
class FloorDivide(Divide):
    """Divide the left value by the right value and round down to the nearest integer."""

    dtype = dt.int64


@public
class Modulus(NumericBinary):
    """Return the remainder after the division of the left value by the right value."""

    dtype = rlz.numeric_like("args", operator.mod)


@public
class Negate(Unary):
    """Negate the value."""

    arg: Value[dt.Numeric | dt.Interval]

    dtype = rlz.dtype_like("arg")


@public
class IsNan(Unary):
    """Check if the value is NaN."""

    arg: Value[dt.Floating]

    dtype = dt.boolean


@public
class IsInf(Unary):
    """Check if the value is infinite."""

    arg: Value[dt.Floating]

    dtype = dt.boolean


@public
class Abs(Unary):
    """Absolute value."""

    arg: SoftNumeric

    dtype = rlz.dtype_like("arg")


@public
class Ceil(Unary):
    """Round up to the nearest integer value greater than or equal to this value."""

    arg: SoftNumeric

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.int64


@public
class Floor(Unary):
    """Round down to the nearest integer value less than or equal to this value."""

    arg: SoftNumeric

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.int64


@public
class Round(Value):
    """Round a value."""

    arg: StrictNumeric
    # TODO(kszucs): the default should be 0 instead of being None
    digits: Optional[Integer] = None

    shape = rlz.shape_like("arg")

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        elif self.digits is None:
            return dt.int64
        else:
            return dt.double


@public
class Clip(Value):
    """Clip a value to a specified range."""

    arg: StrictNumeric
    lower: Optional[StrictNumeric] = None
    upper: Optional[StrictNumeric] = None

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("arg")


@public
class BaseConvert(Value):
    """Convert a number from one base to another."""

    # TODO(kszucs): this should be Integer simply
    arg: Value[dt.Integer | dt.String]
    from_base: Integer
    to_base: Integer

    dtype = dt.string
    shape = rlz.shape_like("args")


@public
class MathUnary(Unary):
    """Base class for unary math operations."""

    arg: SoftNumeric

    @attribute
    def dtype(self):
        return dt.higher_precedence(self.arg.dtype, dt.float64)


class ExpandingMathUnary(MathUnary):
    @attribute
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.float64


@public
class Exp(ExpandingMathUnary):
    """Exponential function."""


@public
class Sign(Unary):
    """Sign of the value."""

    arg: SoftNumeric

    dtype = rlz.dtype_like("arg")


@public
class Sqrt(MathUnary):
    """Square root of the value."""


@public
class Logarithm(MathUnary):
    """Base class for logarithmic operations."""

    arg: StrictNumeric


@public
class Log(Logarithm):
    """Logarithm with a specific base."""

    base: Optional[StrictNumeric] = None


@public
class Ln(Logarithm):
    """Natural logarithm."""


@public
class Log2(Logarithm):
    """Logarithm base 2."""


@public
class Log10(Logarithm):
    """Logarithm base 10."""


@public
class Degrees(ExpandingMathUnary):
    """Converts radians to degrees."""


@public
class Radians(MathUnary):
    """Converts degrees to radians."""


@public
class TrigonometricUnary(MathUnary):
    """Trigonometric base unary."""


@public
class TrigonometricBinary(Binary):
    """Trigonometric base binary."""

    left: SoftNumeric
    right: SoftNumeric

    dtype = dt.float64


@public
class Acos(TrigonometricUnary):
    """Returns the arc cosine of x."""


@public
class Asin(TrigonometricUnary):
    """Returns the arc sine of x."""


@public
class Atan(TrigonometricUnary):
    """Returns the arc tangent of x."""


@public
class Atan2(TrigonometricBinary):
    """Returns the arc tangent of x and y."""


@public
class Cos(TrigonometricUnary):
    """Returns the cosine of x."""


@public
class Cot(TrigonometricUnary):
    """Returns the cotangent of x."""


@public
class Sin(TrigonometricUnary):
    """Returns the sine of x."""


@public
class Tan(TrigonometricUnary):
    """Returns the tangent of x."""


@public
class BitwiseNot(Unary):
    """Bitwise NOT operation."""

    arg: Integer

    dtype = rlz.numeric_like("args", operator.invert)


@public
class BitwiseBinary(Binary):
    """Base class for bitwise binary operations."""

    left: Integer
    right: Integer


@public
class BitwiseAnd(BitwiseBinary):
    """Bitwise AND operation."""

    dtype = rlz.numeric_like("args", operator.and_)


@public
class BitwiseOr(BitwiseBinary):
    """Bitwise OR operation."""

    dtype = rlz.numeric_like("args", operator.or_)


@public
class BitwiseXor(BitwiseBinary):
    """Bitwise XOR operation."""

    dtype = rlz.numeric_like("args", operator.xor)


@public
class BitwiseLeftShift(BitwiseBinary):
    """Bitwise left shift operation."""

    shape = rlz.shape_like("args")
    dtype = dt.int64


@public
class BitwiseRightShift(BitwiseBinary):
    """Bitwise right shift operation."""

    shape = rlz.shape_like("args")
    dtype = dt.int64



from __future__ import annotations

import inspect
import string
import textwrap
from functools import partial, reduce
from itertools import takewhile
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.rules as rlz
from xorq.common.exceptions import InvalidDecoratorError
from xorq.vendor.ibis.backends.sql.compilers.base import (
    NULL,
    STAR,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import PostgresType
from xorq.vendor.ibis.backends.sql.dialects import Postgres
from xorq.vendor.ibis.backends.sql.rewrites import split_select_distinct_with_order_by
from xorq.vendor.ibis.util import gen_name


if TYPE_CHECKING:
    from collections.abc import Mapping

    import xorq.vendor.ibis.expr.types as ir


def _verify_source_line(func_name: str, line: str):
    if line.startswith("@"):
        raise InvalidDecoratorError(func_name, line)
    return line


class PostgresUDFNode(ops.Value):
    shape = rlz.shape_like("args")


class PostgresCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Postgres
    type_mapper = PostgresType
    post_rewrites = (split_select_distinct_with_order_by,)

    agg = AggGen(supports_filter=True, supports_order_by=True)

    NAN = sge.Literal.number("'NaN'::double precision")
    POS_INF = sge.Literal.number("'Inf'::double precision")
    NEG_INF = sge.Literal.number("'-Inf'::double precision")

    UNSUPPORTED_OPS = (
        ops.RowID,
        ops.TimeDelta,
        ops.ArrayFlatten,
    )

    SIMPLE_OPS = {
        ops.Arbitrary: "first",  # could use any_value for postgres>=16
        ops.ArrayRemove: "array_remove",
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.GeoArea: "st_area",
        ops.GeoAsBinary: "st_asbinary",
        ops.GeoAsEWKB: "st_asewkb",
        ops.GeoAsEWKT: "st_asewkt",
        ops.GeoAsText: "st_astext",
        ops.GeoAzimuth: "st_azimuth",
        ops.GeoBuffer: "st_buffer",
        ops.GeoCentroid: "st_centroid",
        ops.GeoContains: "st_contains",
        ops.GeoContainsProperly: "st_contains",
        ops.GeoCoveredBy: "st_coveredby",
        ops.GeoCovers: "st_covers",
        ops.GeoCrosses: "st_crosses",
        ops.GeoDFullyWithin: "st_dfullywithin",
        ops.GeoDWithin: "st_dwithin",
        ops.GeoDifference: "st_difference",
        ops.GeoDisjoint: "st_disjoint",
        ops.GeoDistance: "st_distance",
        ops.GeoEndPoint: "st_endpoint",
        ops.GeoEnvelope: "st_envelope",
        ops.GeoEquals: "st_equals",
        ops.GeoGeometryN: "st_geometryn",
        ops.GeoGeometryType: "st_geometrytype",
        ops.GeoIntersection: "st_intersection",
        ops.GeoIntersects: "st_intersects",
        ops.GeoIsValid: "st_isvalid",
        ops.GeoLength: "st_length",
        ops.GeoLineLocatePoint: "st_linelocatepoint",
        ops.GeoLineMerge: "st_linemerge",
        ops.GeoLineSubstring: "st_linesubstring",
        ops.GeoNPoints: "st_npoints",
        ops.GeoOrderingEquals: "st_orderingequals",
        ops.GeoOverlaps: "st_overlaps",
        ops.GeoPerimeter: "st_perimeter",
        ops.GeoSRID: "st_srid",
        ops.GeoSetSRID: "st_setsrid",
        ops.GeoSimplify: "st_simplify",
        ops.GeoStartPoint: "st_startpoint",
        ops.GeoTouches: "st_touches",
        ops.GeoTransform: "st_transform",
        ops.GeoUnaryUnion: "st_union",
        ops.GeoUnion: "st_union",
        ops.GeoWithin: "st_within",
        ops.GeoX: "st_x",
        ops.GeoY: "st_y",
        ops.MapContains: "exist",
        ops.MapKeys: "akeys",
        ops.MapValues: "avals",
        ops.RegexSearch: "regexp_like",
        ops.TimeFromHMS: "make_time",
    }

    def to_sqlglot(
        self,
        expr: ir.Expr,
        *,
        limit: str | None = None,
        params: Mapping[ir.Expr, Any] | None = None,
    ):
        table_expr = expr.as_table()
        geocols = table_expr.schema().geospatial
        conversions = {name: table_expr[name].as_ewkb() for name in geocols}

        if conversions:
            table_expr = table_expr.mutate(**conversions)
        return super().to_sqlglot(table_expr, limit=limit, params=params)

    def _compile_python_udf(self, udf_node: ops.ScalarUDF):
        config = udf_node.__config__
        func = udf_node.__func__
        func_name = func.__name__

        lines, _ = inspect.getsourcelines(func)
        iter_lines = iter(lines)

        function_premable_lines = list(
            takewhile(lambda line: not line.lstrip().startswith("def "), iter_lines)
        )

        if len(function_premable_lines) > 1:
            raise InvalidDecoratorError(
                name=func_name, lines="".join(function_premable_lines)
            )

        source = textwrap.dedent(
            "".join(map(partial(_verify_source_line, func_name), iter_lines))
        ).strip()

        type_mapper = self.type_mapper
        argnames = udf_node.argnames
        return """\
    CREATE OR REPLACE FUNCTION {ident}({signature})
    RETURNS {return_type}
    LANGUAGE {language}
    AS $$
    {source}
    return {name}({args})
    $$""".format(
            name=type(udf_node).__name__,
            ident=self.__sql_name__(udf_node),
            signature=", ".join(
                f"{argname} {type_mapper.to_string(arg.dtype)}"
                for argname, arg in zip(argnames, udf_node.args)
            ),
            return_type=type_mapper.to_string(udf_node.dtype),
            language=config.get("language", "plpython3u"),
            source=source,
            args=", ".join(argnames),
        )

    def visit_RandomUUID(self, op, **kwargs):
        return self.f.gen_random_uuid()

    def visit_Mode(self, op, *, arg, where):
        expr = self.f.mode()
        expr = sge.WithinGroup(
            this=expr,
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )
        if where is not None:
            expr = sge.Filter(this=expr, expression=sge.Where(this=where))
        return expr

    def visit_ArgMinMax(self, op, *, arg, key, where, desc: bool):
        conditions = [arg.is_(sg.not_(NULL)), key.is_(sg.not_(NULL))]

        if where is not None:
            conditions.append(where)

        agg = self.agg.array_agg(
            sge.Ordered(this=sge.Order(this=arg, expressions=[key]), desc=desc),
            where=sg.and_(*conditions),
        )
        return sge.paren(agg, copy=False)[0]

    def visit_ArgMin(self, op, *, arg, key, where):
        return self.visit_ArgMinMax(op, arg=arg, key=key, where=where, desc=False)

    def visit_ArgMax(self, op, *, arg, key, where):
        return self.visit_ArgMinMax(op, arg=arg, key=key, where=where, desc=True)

    def visit_Sum(self, op, *, arg, where):
        arg = (
            self.cast(self.cast(arg, dt.int32), op.dtype)
            if op.arg.dtype.is_boolean()
            else arg
        )
        return self.agg.sum(arg, where=where)

    def visit_IsNan(self, op, *, arg):
        return arg.eq(self.cast(sge.convert("NaN"), op.arg.dtype))

    def visit_IsInf(self, op, *, arg):
        return arg.isin(self.POS_INF, self.NEG_INF)

    def visit_CountDistinctStar(self, op, *, where, arg):
        # use a tuple because postgres doesn't accept COUNT(DISTINCT a, b, c, ...)
        #
        # this turns the expression into COUNT(DISTINCT ROW(a, b, c, ...))
        row = sge.Tuple(
            expressions=list(
                map(partial(sg.column, quoted=self.quoted), op.arg.schema.keys())
            )
        )
        return self.agg.count(sge.Distinct(expressions=[row]), where=where)

    def visit_Quantile(self, op, *, arg, quantile, where):
        suffix = "cont" if op.arg.dtype.is_numeric() else "disc"
        funcname = f"percentile_{suffix}"
        expr = sge.WithinGroup(
            this=self.f[funcname](quantile),
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )
        if where is not None:
            expr = sge.Filter(this=expr, expression=sge.Where(this=where))
        return expr

    visit_MultiQuantile = visit_Quantile
    visit_ApproxQuantile = visit_Quantile
    visit_ApproxMultiQuantile = visit_Quantile

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                f"{self.dialect} only implements `pop` correlation coefficient"
            )

        # TODO: rewrite rule?
        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_ApproxMedian(self, op, *, arg, where):
        return self.visit_Median(op, arg=arg, where=where)

    def visit_Median(self, op, *, arg, where):
        return self.visit_Quantile(op, arg=arg, quantile=sge.convert(0.5), where=where)

    def visit_ApproxCountDistinct(self, op, *, arg, where):
        return self.agg.count(sge.Distinct(expressions=[arg]), where=where)

    def visit_Range(self, op, *, start, stop, step):
        def zero_value(dtype):
            if dtype.is_interval():
                return self.f.make_interval()
            return 0

        def interval_sign(v):
            zero = self.f.make_interval()
            return sge.Case(
                ifs=[
                    self.if_(v.eq(zero), 0),
                    self.if_(v < zero, -1),
                    self.if_(v > zero, 1),
                ],
                default=NULL,
            )

        def _sign(value, dtype):
            if dtype.is_interval():
                return interval_sign(value)
            return self.f.sign(value)

        step_dtype = op.step.dtype
        return self.if_(
            sg.and_(
                self.f.nullif(step, zero_value(step_dtype)).is_(sg.not_(NULL)),
                _sign(step, step_dtype).eq(_sign(stop - start, step_dtype)),
            ),
            self.f.array_remove(
                self.f.array(
                    sg.select(STAR).from_(self.f.generate_series(start, stop, step))
                ),
                stop,
            ),
            self.cast(self.f.array(), op.dtype),
        )

    visit_IntegerRange = visit_TimestampRange = visit_Range

    def visit_StringConcat(self, op, *, arg):
        return reduce(lambda x, y: sge.DPipe(this=x, expression=y), arg)

    def visit_ArrayConcat(self, op, *, arg):
        return reduce(self.f.array_cat, map(partial(self.cast, to=op.dtype), arg))

    def visit_ArrayContains(self, op, *, arg, other):
        arg_dtype = op.arg.dtype
        # ArrayContainsAll introduced in 24, keep backcompat if it doesn't exist
        cls = getattr(sge, "ArrayContainsAll", sge.ArrayContains)
        return cls(
            this=self.cast(arg, arg_dtype),
            expression=self.f.array(self.cast(other, arg_dtype.value_type)),
        )

    def visit_ArrayFilter(self, op, *, arg, body, param):
        return self.f.array(
            sg.select(sg.column(param, quoted=self.quoted))
            .from_(sge.Unnest(expressions=[arg], alias=param))
            .where(body)
        )

    def visit_ArrayMap(self, op, *, arg, body, param):
        return self.f.array(
            sg.select(body).from_(sge.Unnest(expressions=[arg], alias=param))
        )

    def visit_ArrayPosition(self, op, *, arg, other):
        t = sge.Unnest(expressions=[arg], alias="value", offset=True)
        idx = sg.column("ordinality")
        value = sg.column("value")
        return self.f.coalesce(
            sg.select(idx).from_(t).where(value.eq(other)).limit(1).subquery(), 0
        )

    def visit_ArraySort(self, op, *, arg):
        return self.f.array(
            sg.select("x").from_(sge.Unnest(expressions=[arg], alias="x")).order_by("x")
        )

    def visit_ArrayRepeat(self, op, *, arg, times):
        i = sg.to_identifier("i")
        length = self.f.cardinality(arg)
        return self.f.array(
            sg.select(arg[i % length + 1]).from_(
                self.f.generate_series(0, length * times - 1).as_(i.name)
            )
        )

    def visit_ArrayDistinct(self, op, *, arg):
        return self.if_(
            arg.is_(NULL), NULL, self.f.array(sg.select(self.f.explode(arg)).distinct())
        )

    def visit_ArrayUnion(self, op, *, left, right):
        return self.f.anon.array(
            sg.union(sg.select(self.f.explode(left)), sg.select(self.f.explode(right)))
        )

    def visit_ArrayIntersect(self, op, *, left, right):
        return self.f.anon.array(
            sg.intersect(
                sg.select(self.f.explode(left)), sg.select(self.f.explode(right))
            )
        )

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.array_agg(arg, where=where, order_by=order_by)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the postgres backend"
            )
        return self.agg.first(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the postgres backend"
            )
        return self.agg.last(arg, where=where, order_by=order_by)

    def visit_Log2(self, op, *, arg):
        return self.cast(
            self.f.log(
                self.cast(sge.convert(2), dt.decimal),
                arg if op.arg.dtype.is_decimal() else self.cast(arg, dt.decimal),
            ),
            op.dtype,
        )

    def visit_Log(self, op, *, arg, base):
        if base is not None:
            if not op.base.dtype.is_decimal():
                base = self.cast(base, dt.decimal)
        else:
            base = self.cast(sge.convert(self.f.exp(1)), dt.decimal)

        if not op.arg.dtype.is_decimal():
            arg = self.cast(arg, dt.decimal)
        return self.cast(self.f.log(base, arg), op.dtype)

    def visit_StructField(self, op, *, arg, field):
        idx = op.arg.dtype.names.index(field) + 1
        # postgres doesn't have anonymous structs :(
        #
        # this works around ibis not having a way to tell sqlglot to transform
        # an exploded array(row) into the equivalent unnest(t) _ (col1, ..., colN)
        # element
        #
        # but also postgres should really support anonymous structs
        return self.cast(
            self.f.jsonb_extract_path(self.f.to_jsonb(arg), f"f{idx:d}"), op.dtype
        )

    def json_typeof(self, op, arg):
        b = "b" * op.arg.dtype.binary
        return self.f[f"json{b}_typeof"](arg)

    def json_extract_path_text(self, op, arg, *rest):
        b = "b" * op.arg.dtype.binary
        return self.f[f"json{b}_extract_path_text"](
            arg,
            *rest,
            # this is apparently how you pass in no additional arguments to
            # a variadic function, see the "Variadic Function Resolution"
            # section in
            # https://www.postgresql.org/docs/current/typeconv-func.html
            sge.Var(this="VARIADIC ARRAY[]::TEXT[]"),
        )

    def visit_UnwrapJSONString(self, op, *, arg):
        return self.if_(
            self.json_typeof(op, arg).eq(sge.convert("string")),
            self.json_extract_path_text(op, arg),
            NULL,
        )

    def visit_UnwrapJSONInt64(self, op, *, arg):
        text = self.json_extract_path_text(op, arg)
        return self.if_(
            self.json_typeof(op, arg).eq(sge.convert("number")),
            self.cast(
                self.if_(self.f.regexp_like(text, r"^\d+$", "g"), text, NULL), op.dtype
            ),
            NULL,
        )

    def visit_UnwrapJSONFloat64(self, op, *, arg):
        text = self.json_extract_path_text(op, arg)
        return self.if_(
            self.json_typeof(op, arg).eq(sge.convert("number")),
            self.cast(text, op.dtype),
            NULL,
        )

    def visit_UnwrapJSONBoolean(self, op, *, arg):
        return self.if_(
            self.json_typeof(op, arg).eq(sge.convert("boolean")),
            self.cast(self.json_extract_path_text(op, arg), op.dtype),
            NULL,
        )

    def visit_JSONGetItem(self, op, *, arg, index):
        if op.arg.dtype.binary:
            return self.f.jsonb_extract_path(arg, self.cast(index, dt.string))
        return super().visit_JSONGetItem(op, arg=arg, index=index)

    def visit_StructColumn(self, op, *, names, values):
        return self.f.row(*map(self.cast, values, op.dtype.types))

    def visit_ToJSONArray(self, op, *, arg):
        b = "b" * op.arg.dtype.binary
        return self.if_(
            self.json_typeof(op, arg).eq(sge.convert("array")),
            self.f.array(sg.select(STAR).from_(self.f[f"json{b}_array_elements"](arg))),
            NULL,
        )

    def visit_Map(self, op, *, keys, values):
        # map(["a", "b"], NULL) results in {"a": NULL, "b": NULL} in regular postgres,
        # so we need to modify it to return NULL instead
        regular = self.f.map(keys, values)
        return self.if_(values.is_(NULL), NULL, regular)

    def visit_MapLength(self, op, *, arg):
        return self.f.cardinality(self.f.akeys(arg))

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(
            self.f.exist(arg, key),
            self.f.jsonb_extract_path_text(self.f.to_jsonb(arg), key),
            default,
        )

    def visit_MapMerge(self, op, *, left, right):
        return sge.DPipe(this=left, expression=right)

    def visit_TypeOf(self, op, *, arg):
        typ = self.cast(self.f.pg_typeof(arg), dt.string)
        return self.if_(
            typ.eq(sge.convert("unknown")),
            "null" if op.arg.dtype.is_null() else "text",
            typ,
        )

    def visit_Round(self, op, *, arg, digits):
        if digits is None:
            return self.f.round(arg)

        result = self.f.round(self.cast(arg, dt.decimal), digits)
        if op.arg.dtype.is_decimal():
            return result
        return self.cast(result, dt.float64)

    def visit_Modulus(self, op, *, left, right):
        # postgres doesn't allow modulus of double precision values, so upcast and
        # then downcast later if necessary
        if not op.dtype.is_integer():
            left = self.cast(left, dt.decimal)
            right = self.cast(right, dt.decimal)

        result = left % right
        if op.dtype.is_float64():
            return self.cast(result, dt.float64)
        else:
            return result

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        pattern = self.f.concat("(", pattern, ")")
        matches = self.f.regexp_match(arg, pattern)
        return self.if_(arg.rlike(pattern), sge.paren(matches, copy=False)[index], NULL)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.coalesce(
            self.f.array_position(self.f.array(*values), needle),
            0,
        )

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.strpos(haystack, needle) > 0

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.length(end)).eq(end)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.cast("".join(map(r"\x{:0>2x}".format, value)), dt.binary)
        elif dtype.is_time():
            to_int32 = partial(self.cast, to=dt.int32)
            to_float64 = partial(self.cast, to=dt.float64)

            return self.f.make_time(
                to_int32(value.hour),
                to_int32(value.minute),
                to_float64(value.second + value.microsecond / 1e6),
            )
        elif dtype.is_json():
            return self.cast(value, dt.json)
        return None

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        to_int32 = partial(self.cast, to=dt.int32)
        return self.f.make_timestamp(
            to_int32(year),
            to_int32(month),
            to_int32(day),
            to_int32(hours),
            to_int32(minutes),
            self.cast(seconds, dt.float64),
        )

    def visit_DateFromYMD(self, op, *, year, month, day):
        to_int32 = partial(self.cast, to=dt.int32)
        return self.f.datefromparts(to_int32(year), to_int32(month), to_int32(day))

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        origin = self.f.make_timestamp(
            *map(partial(self.cast, to=dt.int32), (1970, 1, 1, 0, 0, 0))
        )

        if offset is not None:
            origin += offset

        return self.f.date_bin(interval, arg, origin)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return self.cast(self.f.extract("dow", arg) + 6, dt.int16) % 7

    def visit_DayOfWeekName(self, op, *, arg):
        return self.f.trim(self.f.to_char(arg, "Day"), string.whitespace)

    def visit_ExtractSecond(self, op, *, arg):
        return self.cast(self.f.floor(self.f.extract("second", arg)), op.dtype)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.cast(
            self.f.floor(self.f.extract("millisecond", arg)) % 1_000, op.dtype
        )

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.extract("microsecond", arg) % 1_000_000

    def visit_ExtractDayOfYear(self, op, *, arg):
        return self.f.extract("doy", arg)

    def visit_ExtractWeekOfYear(self, op, *, arg):
        return self.f.extract("week", arg)

    def visit_ExtractIsoYear(self, op, *, arg):
        return self.f.extract("isoyear", arg)

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.extract("epoch", arg)

    def visit_ArrayIndex(self, op, *, arg, index):
        index = self.if_(index < 0, self.f.cardinality(arg) + index, index)
        return sge.paren(arg, copy=False)[index]

    def visit_ArraySlice(self, op, *, arg, start, stop):
        neg_to_pos_index = lambda n, index: self.if_(index < 0, n + index, index)  # noqa: E731

        arg_length = self.f.cardinality(arg)

        if start is None:
            start = 0
        else:
            start = self.f.least(arg_length, neg_to_pos_index(arg_length, start))

        if stop is None:
            stop = arg_length
        else:
            stop = neg_to_pos_index(arg_length, stop)

        slice_expr = sge.Slice(this=start + 1, expression=stop)
        return sge.paren(arg, copy=False)[slice_expr]

    def _make_interval(self, arg, unit):
        plural = unit.plural
        if plural == "minutes":
            plural = "mins"
            arg = self.cast(arg, dt.int32)
        elif plural == "seconds":
            plural = "secs"
            arg = self.cast(arg, dt.float64)
        elif plural == "milliseconds":
            plural = "secs"
            arg /= 1e3
        elif plural == "microseconds":
            plural = "secs"
            arg /= 1e6
        elif plural == "nanoseconds":
            plural = "secs"
            arg /= 1e9
        else:
            arg = self.cast(arg, dt.int32)

        key = sg.to_identifier(plural)

        return self.f.make_interval(sge.Kwarg(this=key, expression=arg))

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype

        if from_.is_timestamp() and to.is_integer():
            return self.f.extract("epoch", arg)
        elif from_.is_integer() and to.is_timestamp():
            arg = self.f.to_timestamp(arg)
            if (timezone := to.timezone) is not None:
                arg = self.f.timezone(timezone, arg)
            return arg
        elif from_.is_string() and to.is_binary():
            # Postgres and Python use the words "decode" and "encode" in
            # opposite ways, sweet!
            return self.f.decode(arg, "escape")
        elif from_.is_binary() and to.is_string():
            return self.f.encode(arg, "escape")

        return super().visit_Cast(op, arg=arg, to=to)

    visit_TryCast = visit_Cast

    def visit_Hash(self, op, *, arg):
        arg_dtype = op.arg.dtype

        if arg_dtype.is_int16():
            return self.f.hashint2extended(arg, 0)
        elif arg_dtype.is_int32():
            return self.f.hashint4extended(arg, 0)
        elif arg_dtype.is_int64():
            return self.f.hashint8extended(arg, 0)
        elif arg_dtype.is_float32():
            return self.f.hashfloat4extended(arg, 0)
        elif arg_dtype.is_float64():
            return self.f.hashfloat8extended(arg, 0)
        elif arg_dtype.is_string():
            return self.f.hashtextextended(arg, 0)
        elif arg_dtype.is_macaddr():
            return self.f.hashmacaddr8extended(arg, 0)

        raise com.UnsupportedOperationError(
            f"Hash({arg_dtype!r}) operation is not supported in the "
            f"{self.dialect} backend"
        )

    def visit_TableUnnest(
        self,
        op,
        *,
        parent,
        column,
        column_name: str,
        offset: str | None,
        keep_empty: bool,
    ):
        quoted = self.quoted

        column_alias = sg.to_identifier(gen_name("table_unnest_column"), quoted=quoted)

        parent_alias = parent.alias_or_name

        parent_schema = op.parent.schema
        overlaps_with_parent = column_name in parent_schema
        computed_column = column_alias.as_(column_name, quoted=quoted)

        selcols = []

        if overlaps_with_parent:
            column_alias_or_name = column.alias_or_name
            selcols.extend(
                sg.column(col, table=parent_alias, quoted=quoted)
                if col != column_alias_or_name
                else computed_column
                for col in parent_schema.names
            )
        else:
            selcols.append(
                sge.Column(
                    this=STAR, table=sg.to_identifier(parent_alias, quoted=quoted)
                )
            )
            selcols.append(computed_column)

        if offset is not None:
            offset_name = offset
            offset = sg.to_identifier(offset_name, quoted=quoted)
            selcols.append((offset - 1).as_(offset_name, quoted=quoted))

        unnest = sge.Unnest(
            expressions=[column],
            alias=sge.TableAlias(
                this=sg.to_identifier(gen_name("table_unnest"), quoted=quoted),
                columns=[column_alias],
            ),
            offset=offset,
        )

        return (
            sg.select(*selcols)
            .from_(parent)
            .join(
                unnest,
                on=None if not keep_empty else sge.convert(True),
                join_type="CROSS" if not keep_empty else "LEFT",
            )
        )

    def _unnest(self, expression, *, as_):
        alias = sge.TableAlias(columns=[sg.to_identifier(as_)])
        return sge.Unnest(expressions=[expression], alias=alias)

    def _array_reduction(self, *, arg, reduction):
        name = sg.to_identifier(gen_name(f"pg_arr_{reduction}"))
        return (
            sg.select(self.f[reduction](name))
            .from_(self._unnest(arg, as_=name))
            .subquery()
        )

    def visit_ArrayMin(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="min")

    def visit_ArrayMax(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="max")

    def visit_ArraySum(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="sum")

    def visit_ArrayMean(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="avg")

    def visit_ArrayAny(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="bool_or")

    def visit_ArrayAll(self, op, *, arg):
        return self._array_reduction(arg=arg, reduction="bool_and")


compiler = PostgresCompiler()



import os
from pathlib import Path

import pytest

from xorq.common.utils.env_utils import (
    EnvConfigable,
    parse_env_file,
)


def test_subclass_from_kwargs(monkeypatch):
    exists_name, missing_name = "X", "SHOULDNT_EXIST"
    monkeypatch.setenv(exists_name, "1")
    monkeypatch.delenv(missing_name, raising=False)

    EnvConfig = EnvConfigable.subclass_from_kwargs(exists_name, missing_name)
    env_config = EnvConfig.from_env()
    assert env_config[exists_name] == "1"
    assert env_config[missing_name] == ""

    # env vars take priority
    env_config = EnvConfig.from_env(**{exists_name: 2})
    assert env_config[exists_name] == "1"
    assert env_config[missing_name] == ""

    # X is an env var so will be passed as a string but must be an int
    EnvConfig = EnvConfigable.subclass_from_kwargs(missing_name, **{exists_name: 1})
    with pytest.raises(TypeError):
        env_config = EnvConfig.from_env()

    # so when using from_env, non-string fields must not exist in the env
    monkeypatch.delenv(exists_name)
    EnvConfig = EnvConfigable.subclass_from_kwargs(missing_name, **{exists_name: 1})
    env_config = EnvConfig.from_env()
    assert env_config[exists_name] == 1
    assert env_config[missing_name] == ""


def make_EnvConfig_from_content(tmp_path, content):
    env_file = Path(tmp_path).joinpath(".env")
    env_file.write_text(content)
    EnvConfig = EnvConfigable.subclass_from_env_file(env_file)
    return EnvConfig


def test_subclass_from_env_file(monkeypatch, tmp_path):
    EnvConfig = make_EnvConfig_from_content(
        tmp_path,
        "\n".join(
            (
                "export X=2",
                "Y=3",
                "a line that does not match, perhaps we should raise?",
            )
        ),
    )

    env_config = EnvConfig()
    assert env_config.X == "2"
    assert env_config.Y == "3"

    env_config = EnvConfig.from_env()
    assert env_config.X == "2"
    assert env_config.Y == "3"

    monkeypatch.setitem(os.environ, "X", "1")
    env_config = EnvConfig.from_env()
    assert env_config.X == "1"
    assert env_config.Y == "3"


def test_single_line_env(tmp_path):
    EnvConfig = make_EnvConfig_from_content(
        tmp_path,
        "x=1",
    )
    env_config = EnvConfig()
    assert env_config.x == "1"


def test_parse_multiline_env_vars(fixture_dir):
    multiline_env_vars_path = fixture_dir.joinpath("multiline_env_vars.env")
    dct = parse_env_file(multiline_env_vars_path)
    actual = {k: len(v.split("\n")) for k, v in dct.items()}
    expected = {
        k: int(v)
        for k, v in (el.split("=") for el in dct["names_to_numlines"].split(","))
    }
    assert actual == expected


def test_clone(monkeypatch):
    initial = {
        "a": "one",
        "b": 2,
        "c": 3.0,
    }
    modifications = {
        "a": "zero",
    }
    bad_modifications = {
        "b": 2.0,
    }

    for name in initial:
        monkeypatch.delenv(name, raising=False)

    EnvConfig = EnvConfigable.subclass_from_kwargs(**initial)
    env_config = EnvConfig()

    cloned = env_config.clone(**modifications)
    assert cloned != env_config
    assert cloned.__getstate__() == env_config.__getstate__() | modifications

    with pytest.raises(TypeError, match="must be <class.*got.*that is a <class"):
        env_config.clone(**bad_modifications)



import itertools
import json
from pathlib import Path

import dask
import toolz
import yaml
from attr import evolve, field, frozen
from attr.validators import (
    instance_of,
    optional,
)

from xorq.common.utils.env_utils import compiled_env_var_substitution_re
from xorq.common.utils.inspect_utils import get_arguments
from xorq.loader import _load_entry_points, load_backend
from xorq.vendor.ibis.config import options


@frozen
class Profiles:
    """A collection interface for managing database connection profiles.

    The Profiles class provides a centralized way to access, manage, and retrieve
    Profile objects saved on the filesystem. It handles directory management,
    profile lookup, and provides a dictionary-like and attribute-like interface
    for accessing saved profiles.

    Profiles are immutable (frozen) to ensure thread safety and prevent accidental
    modification after creation.

    Attributes
    ----------
    profile_dir : Path
        Directory where profile files are stored. Defaults to xo.options.profiles.profile_dir.
        Created automatically if it doesn't exist.

    Examples
    --------
    Creating a Profiles collection:
    >>> import xorq.api as xo
    >>> from xorq.vendor.ibis.backends.profiles import Profile, Profiles
    >>> profiles = Profiles()
    >>> # Or with a custom directory
    >>> from pathlib import Path
    >>> custom_profiles = Profiles(profile_dir=Path('/path/to/profiles'))

    Accessing profiles by name (attribute-style):
    >>> Profile.from_con(xo.connect()).save(alias='example_dev')
    >>> postgres_profile = profiles.example_dev

    Accessing profiles by name (dictionary-style):

    >>> again_example_dev = profiles['example_dev']

    Getting a profile explicitly:

    >>> profile = profiles.get('example_dev')

    Listing available profiles:

    >>> profiles.list()
    ('example_dev', 'feda6956a9ca4d2bda0fbc8e775042c3_1')


    Notes
    -----
    The Profiles class supports tab-completion in IPython and Jupyter environments,
    making it easy to discover available profiles interactively.

    The directory structure uses YAML files for profile storage, with optional
    symbolic links for aliased profiles.



    See Also
    --------
    Profile : Individual connection profile class
    Profile.save : Save a profile to disk
    Profile.load : Load a profile from disk
    """

    profile_dir = field(validator=optional(instance_of(Path)), default=None)

    def __attrs_post_init__(self):
        if self.profile_dir is None:
            # defer setting to pick up in-process changes to options
            object.__setattr__(self, "profile_dir", options.profiles.profile_dir)
        if not self.profile_dir.exists():
            self.profile_dir.mkdir(exist_ok=True, parents=True)

    def get(self, name):
        return Profile.load(name, profile_dir=self.profile_dir)

    def __getattr__(self, stem):
        try:
            return self.get(
                next(el.name for el in self.profile_dir.iterdir() if el.stem == stem)
            )
        except Exception:
            return object.__getattribute__(self, stem)

    def __getitem__(self, stem):
        return self.get(
            next(el.name for el in self.profile_dir.iterdir() if el.stem == stem)
        )

    def __dir__(self):
        return tuple(el for el in self.list() if el.isidentifier())

    def list(self):
        return tuple(el.stem for el in self.profile_dir.iterdir())

    def _ipython_key_completions_(self):
        return self.list()


@frozen
class Profile:
    """A representation of a database connection profile that can be saved and loaded.

    The Profile class encapsulates all the information needed to establish a database
    connection, including connection type and parameters. It supports serialization to
    and deserialization from YAML, environment variable substitution, and security
    checks to prevent sensitive information from being stored in plain text.

    Profiles are immutable (frozen) to ensure thread safety and prevent accidental
    modification after creation.

    Attributes
    ----------
    con_name : str
        Name of the connection backend (e.g., 'postgres', 'snowflake')
    kwargs_tuple : tuple
        Connection parameters as a tuple of (key, value) pairs
    idx : int
        Unique identifier for this profile instance, auto-generated if not provided

    Examples
    --------
    Creating a profile with environment variables for sensitive information:

    >>> profile = Profile(
    ...     con_name='postgres',
    ...     kwargs_tuple=(
    ...         ('host', '${POSTGRES_HOST}'),
    ...         ('port', 5432),
    ...         ('database', 'mydb'),
    ...         ('user', '${POSTGRES_USER}'),
    ...         ('password', '${POSTGRES_PASSWORD}'),
    ...     )
    ... )

    Saving a profile:

    >>> profile.save(alias='my_postgres')

    Loading a profile:

    >>> loaded_profile = Profile.load('my_postgres')

    Creating a connection from a profile:

    >>> connection = profile.get_con()

    Creating a profile from an existing connection:

    >>> new_profile = Profile.from_con(connection)

    Notes
    -----
    Sensitive information like passwords should be stored as environment variable
    references (${VAR} or $VAR) to prevent security risks. The `save` method will
    raise a ValueError if sensitive data is not stored as environment variables.
    """

    con_name = field(validator=instance_of(str))
    kwargs_tuple = field(validator=instance_of(tuple))
    idx = field(validator=instance_of(int), factory=itertools.count().__next__)

    @con_name.validator
    def validate_con_name(self, attr, value):
        assert next(
            (ep for ep in _load_entry_points() if ep.name == value),
            None,
        )

    def __attrs_post_init__(self):
        # Sort kwargs_tuple after initialization
        object.__setattr__(self, "kwargs_tuple", tuple(sorted(self.kwargs_tuple)))

    @property
    def kwargs_dict(self):
        return dict(self.kwargs_tuple)

    @property
    def hash_name(self):
        dask_hash = dask.base.tokenize(toolz.dissoc(self.as_dict(), "idx"))
        return f"{dask_hash}_{self.idx}"

    def get_con(self, **kwargs):
        """Create a connection using this profile's parameters."""
        _kwargs = dict(self.kwargs_tuple) | kwargs
        connect = getattr(load_backend(self.con_name), "connect")
        con = connect(**_kwargs)
        return con

    def clone(self, idx=None, **kwargs):
        idx = idx if idx is not None else self.idx
        kwargs_tuple = tuple((dict(self.kwargs_tuple) | kwargs).items())

        return evolve(self, idx=idx, kwargs_tuple=kwargs_tuple)

    def as_dict(self):
        return {
            name: getattr(self, name)
            for name in (attr.name for attr in self.__attrs_attrs__)
        }

    def as_json(self):
        return json.dumps(self.as_dict())

    def as_yaml(self):
        return yaml.safe_dump(
            {
                "con_name": self.con_name,
                "kwargs_dict": self.kwargs_dict,
                "idx": self.idx,
            }
        )

    def save(self, profile_dir=None, alias=None, clobber=False, check_secrets=True):
        """Save this profile to disk as a YAML file.

        This method serializes the profile to YAML format and writes it to a file in the
        specified profile directory. The filename is automatically generated based on a hash
        of the profile's content, ensuring uniqueness. Optionally creates a symlink with a
        user-friendly alias name for easier access.

        Before saving, the method can check for exposed secret keys (like passwords) and
        ensure they're stored as environment variable references rather than plain text.

        Parameters
        ----------
        profile_dir : Path, optional
            Directory where the profile will be saved. If None, uses the default directory
            from xo.options.profiles.profile_dir.
        alias : str, optional
            If provided, creates a symbolic link with this name pointing to the saved profile
            file. Useful for giving profiles memorable names.
        clobber : bool, default False
            If True, overwrites existing files with the same name. If False and a file with
            the same name exists, returns the existing file path without overwriting.
        check_secrets : bool, default True
            If True, checks that sensitive information (like passwords) is stored as
            environment variable references before saving. If False, skips this check.

        Returns
        -------
        Path
            Path to the saved profile file or alias symlink (if created).

        Raises
        ------
        ValueError
            If check_secrets is True and the profile contains exposed secret keys.
            If clobber is False and an alias path already exists.

        Notes
        -----
        The saved YAML file contains the connection type, parameters, and a unique identifier.
        The file is saved with a hash-based name to ensure uniqueness, but for user-friendly
        access, an alias (symlink) can be created.

        Environment variable references in the profile are saved as-is, not substituted with
        actual values, to maintain security and portability.

        See Also
        --------
        load : Load a previously saved profile
        check_for_exposed_secrets : Check for sensitive information not using env vars

        Examples
        --------
        >>> # Save a profile with default options
        >>> profile.save()


        >>> # Save with an alias name
        >>> profile.save(alias='postgres_dev')


        >>> # Save to a custom directory and overwrite if exists
        >>> from pathlib import Path
        >>> custom_dir = Path('/path/to/profiles')
        >>> profile.save(profile_dir=custom_dir, clobber=True)


        >>> # Save without checking for exposed secrets (not recommended)
        >>> profile.save(check_secrets=False)
        """
        if check_secrets:
            self.check_for_exposed_secrets()
        path = self.get_path(self.hash_name, profile_dir=profile_dir)
        if not path.exists():
            path.write_text(self.as_yaml())
        if alias:
            alias_path = self.get_path(alias, profile_dir=profile_dir)
            if alias_path.exists():
                if not clobber:
                    raise ValueError
                alias_path.unlink()
            alias_path.symlink_to(path)
            return alias_path
        return path

    def check_for_exposed_secrets(self):
        check_for_exposed_secrets(self.con_name, self.kwargs_dict)

    def almost_equals(self, other):
        return self.clone(idx=-1) == other.clone(idx=-1)

    @classmethod
    def get_path(cls, name, profile_dir=None):
        profile_dir = profile_dir or options.profiles.profile_dir
        profile_dir.mkdir(exist_ok=True, parents=True)
        path = profile_dir.joinpath(name).with_suffix(".yaml")
        return path

    @classmethod
    def load(cls, name, profile_dir=None):
        """Load a Profile from disk by name or hash.

        This method retrieves a serialized Profile from the filesystem and deserializes it
        into a Profile object. It handles locating the profile file, parsing the YAML
        content, and reconstructing the Profile with all its original parameters.

        Parameters
        ----------
        name : str
            The name or hash identifier of the profile to load. This can be either an alias
            name created during save or the hash-based filename.
        profile_dir : Path, optional
            Directory containing profile files. If None, uses the default directory from
            xo.options.profiles.profile_dir.

        Returns
        -------
        Profile
            A Profile object with the connection parameters loaded from the file.

        Raises
        ------
        FileNotFoundError
            If no profile with the given name exists in the profile directory.
        ValueError
            If the profile file exists but contains invalid or incomplete data.

        Notes
        -----
        The loaded profile will have the same unique identifier (idx) as when it was saved,
        preserving its identity across serialization cycles. Environment variable references
        in the profile are preserved and not substituted during loading.

        See Also
        --------
        save : Save a Profile to disk
        get_path : Get the filesystem path for a profile name

        Examples
        --------
        >>> # Load a profile by its alias
        >>> profile = Profile.load('postgres_dev')
        >>>
        >>> # Load a profile from a non-default directory
        >>> from pathlib import Path
        >>> custom_dir = Path('/path/to/profiles')
        >>> profile = Profile.load('postgres_dev', profile_dir=custom_dir)
        """
        path = cls.get_path(name, profile_dir=profile_dir)
        env = yaml.safe_load(path.read_text())
        con_name = env.get("con_name")
        kwargs_dict = env.get("kwargs_dict")
        idx = env.get("idx")
        sorted_kwargs = tuple(sorted(kwargs_dict.items()))
        return cls(con_name=con_name, kwargs_tuple=sorted_kwargs, idx=idx)

    @classmethod
    def from_con(cls, con, *args, **kwargs):
        """Create a Profile from a connection, preserving env var references if possible."""

        def get_combined_arguments():
            # these are the env-mapped values
            arguments0 = get_arguments(
                con.do_connect, *con._con_args, **con._con_kwargs
            )
            # these are the "raw" values (if passed)
            arguments1 = toolz.valfilter(
                bool, get_arguments(con.do_connect, *args, **kwargs)
            )
            assert not arguments0.get("args")
            assert not arguments1.get("args")
            arguments = toolz.dissoc(arguments0 | arguments1, "args")
            return arguments

            # If connection already has a profile, return it to preserve env vars

        if hasattr(con, "_profile") and con._profile is not None:
            return con._profile

        if con.name == "xorq_flight":
            return None

        kwargs_name = "config" if con.name == "duckdb" else "kwargs"
        arguments = get_combined_arguments()
        kwargs = toolz.dissoc(arguments, kwargs_name) | arguments.get(kwargs_name, {})

        # Fix port type if needed
        if (
            "port" in kwargs
            and kwargs["port"] is not None
            and isinstance(kwargs["port"], str)
        ):
            kwargs["port"] = int(kwargs["port"])

        return cls(con_name=con.name, kwargs_tuple=tuple(sorted(kwargs.items())))


def check_for_exposed_secrets(con_name: str, kwargs: dict) -> None:
    """Check if profile contains exposed secret keys.

    Raises
    ------
    ValueError
        If profile contains exposed secret keys not using environment variables
    """

    # Define secret keys by connection name
    # TODO: Add more database types as needed
    # maybe user sets this in options
    con_name_to_secret_keys = {
        "postgres": [
            "password",
            "sslcert",
            "sslkey",
            "sslrootcert",
            "sslcrl",
            "options",
            "passfile",
        ],
        "snowflake": [
            "password",
            "user",
            "account",
            "token",
            "private_key",
            "private_key_path",
            "oauth_token",
        ],
        # Add more database types as needed
    }

    relevant_keys = con_name_to_secret_keys.get(
        con_name,
        ["password"],  # default to just password
    )

    exposed_secrets = tuple(
        key
        for key, value in kwargs.items()
        if key in relevant_keys
        and not (
            isinstance(value, str) and compiled_env_var_substitution_re.match(value)
        )
    )
    if exposed_secrets:
        secrets_list = ", ".join(f"'{key}'" for key in exposed_secrets)
        env_var_examples = ", ".join(f"${key} or ${{{key}}}" for key in exposed_secrets)
        raise ValueError(
            f"Profile contains exposed secret keys: {secrets_list}. "
            f"Use environment variables ({env_var_examples}) for these values."
        )



"""Flink Ibis expression to SQL compiler."""

from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import (
    NULL,
    STAR,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import FlinkType
from xorq.vendor.ibis.backends.sql.dialects import Flink
from xorq.vendor.ibis.backends.sql.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_rank,
    exclude_unsupported_window_frame_from_row_number,
)


class FlinkAggGen(AggGen):
    def aggregate(self, compiler, name, *args, where=None, order_by=()):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )

        func = compiler.f[name]
        if where is not None:
            # Flink does support FILTER, but it's broken for:
            #
            # 1. certain aggregates: std/var doesn't return the right result
            # 2. certain kinds of predicates: x IN y doesn't filter the right
            #    values out
            # 3. certain aggregates AND predicates STD(w) FILTER (WHERE x IN Y)
            #    returns an incorrect result
            #
            # One solution is to try `IF(predicate, arg, NULL)`.
            #
            # Unfortunately that won't work without casting the NULL to a
            # specific type.
            #
            # At this point in the Ibis compiler we don't have any of the Ibis
            # operation's type information because we thrown it away. In every
            # other engine Ibis supports the type of a NULL literal is inferred
            # by the engine.
            #
            # Using a CASE statement and leaving out the explicit NULL does the
            # trick for Flink.
            args = tuple(sge.Case(ifs=[sge.If(this=where, true=arg)]) for arg in args)
        return func(*args)


class FlinkCompiler(SQLGlotCompiler):
    quoted = True
    dialect = Flink
    type_mapper = FlinkType

    agg = FlinkAggGen()

    rewrites = (
        exclude_unsupported_window_frame_from_row_number,
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_rank,
        *SQLGlotCompiler.rewrites,
    )

    UNSUPPORTED_OPS = (
        ops.AnalyticVectorizedUDF,
        ops.ApproxMedian,
        ops.ArgMax,
        ops.ArgMin,
        ops.ArrayFlatten,
        ops.ArrayStringJoin,
        ops.Correlation,
        ops.CountDistinctStar,
        ops.Covariance,
        ops.DateDiff,
        ops.ExtractURLField,
        ops.FindInSet,
        ops.IsInf,
        ops.IsNan,
        ops.Levenshtein,
        ops.Median,
        ops.NthValue,
        ops.ReductionVectorizedUDF,
        ops.RegexSplit,
        ops.RowID,
        ops.StringSplit,
        ops.Translate,
    )

    SIMPLE_OPS = {
        ops.All: "min",
        ops.Any: "max",
        ops.ApproxCountDistinct: "approx_count_distinct",
        ops.ArrayDistinct: "array_distinct",
        ops.ArrayLength: "cardinality",
        ops.ArrayPosition: "array_position",
        ops.ArrayRemove: "array_remove",
        ops.ArraySort: "array_sort",
        ops.ArrayUnion: "array_union",
        ops.ExtractDayOfYear: "dayofyear",
        ops.MapKeys: "map_keys",
        ops.MapValues: "map_values",
        ops.Power: "power",
        ops.RegexSearch: "regexp",
        ops.StrRight: "right",
        ops.StringLength: "char_length",
        ops.StringToDate: "to_date",
        ops.StringToTimestamp: "to_timestamp",
        ops.TypeOf: "typeof",
    }

    @property
    def NAN(self):
        raise NotImplementedError("Flink does not support NaN")

    @property
    def POS_INF(self):
        raise NotImplementedError("Flink does not support Infinity")

    NEG_INF = POS_INF

    @staticmethod
    def _generate_groups(groups):
        return groups

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        elif (
            isinstance(getattr(end, "value", None), ops.Cast)
            and end.value.arg.value == 0
            and end.following
        ):
            spec.args["end"] = "CURRENT ROW"
            spec.args["end_side"] = None
        return spec

    def visit_WindowAggregate(
        self,
        op,
        *,
        parent,
        window_type,
        time_col,
        groups,
        metrics,
        window_size,
        window_slide,
        window_offset,
    ):
        if window_type == "tumble":
            assert window_slide is None

        args = [
            self.v[f"TABLE {parent.this.sql(self.dialect)}"],
            # `time_col` has the table _alias_, instead of the table, but it is
            # required to be bound to the table, this happens because of the
            # way we construct the op in the tumble API using bind
            #
            # perhaps there's a better way to deal with this
            self.f.descriptor(time_col.this),
            window_slide,
            window_size,
            window_offset,
        ]

        window_func = getattr(self.f, window_type)

        # create column references to new columns generated by Flink in the output
        window_start = sg.column(
            "window_start", table=parent.alias_or_name, quoted=True
        )
        window_end = sg.column("window_end", table=parent.alias_or_name, quoted=True)

        return (
            sg.select(
                window_start,
                window_end,
                *self._cleanup_names(groups),
                *self._cleanup_names(metrics),
                copy=False,
            )
            .from_(
                self.f.table(window_func(*filter(None, args))).as_(
                    parent.alias_or_name, quoted=True
                )
            )
            .group_by(
                *self._generate_groups([window_start, window_end, *groups.values()])
            )
        )

    def visit_InMemoryTable(self, op, *, name, schema, data):
        # the performance of this is rather terrible
        tuples = data.to_frame().itertuples(index=False)
        quoted = self.quoted
        columns = [sg.column(col, quoted=quoted) for col in schema.names]
        alias = sge.TableAlias(
            this=sg.to_identifier(name, quoted=quoted), columns=columns
        )
        expressions = [
            sge.Tuple(
                expressions=[
                    self.visit_Literal(
                        ops.Literal(col, dtype=dtype), value=col, dtype=dtype
                    )
                    for col, dtype in zip(row, schema.types)
                ]
            )
            for row in tuples
        ]

        expr = sge.Values(expressions=expressions, alias=alias)
        return sg.select(*columns).from_(expr)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            # TODO: is this decode safe?
            return self.cast(value.decode(), dtype)
        elif dtype.is_uuid():
            return sge.convert(str(value))
        elif dtype.is_array():
            value_type = dtype.value_type
            result = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                    )
                    for v in value
                )
            )
            if value:
                return result
            return sge.Cast(this=result, to=self.type_mapper.from_ibis(dtype))
        elif dtype.is_map():
            key_type = dtype.key_type
            value_type = dtype.value_type
            keys = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=key_type), value=v, dtype=key_type
                    )
                    for v in value.keys()
                )
            )
            values = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                    )
                    for v in value.values()
                )
            )
            return self.cast(self.f.map_from_arrays(keys, values), dtype)
        elif dtype.is_timestamp():
            return self.cast(
                value.replace(tzinfo=None).isoformat(sep=" ", timespec="microseconds"),
                dtype,
            )
        elif dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_time():
            return self.cast(value.isoformat(timespec="microseconds"), dtype)
        return None

    def visit_ArrayIndex(self, op, *, arg, index):
        return sge.Bracket(this=arg, expressions=[index + 1])

    def visit_Xor(self, op, *, left, right):
        return sg.or_(sg.and_(left, sg.not_(right)), sg.and_(sg.not_(left), right))

    def visit_Literal(self, op, *, value, dtype):
        if value is None:
            assert dtype.nullable, "dtype is not nullable but value is None"
            if not dtype.is_null():
                return self.cast(NULL, dtype)
            return NULL
        return super().visit_Literal(op, value=value, dtype=dtype)

    def visit_MapGet(self, op, *, arg, key, default):
        if default is NULL:
            default = self.cast(default, op.dtype)
        return self.f.coalesce(arg[self.cast(key, op.arg.dtype.key_type)], default)

    def visit_ArraySlice(self, op, *, arg, start, stop):
        args = [arg, self.if_(start >= 0, start + 1, start)]

        if stop is not None:
            args.append(
                self.if_(stop >= 0, stop, self.f.cardinality(arg) - self.f.abs(stop))
            )

        return self.f.array_slice(*args)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the flink backend"
            )
        return self.agg.first_value(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the flink backend"
            )
        return self.agg.last_value(arg, where=where, order_by=order_by)

    def visit_Not(self, op, *, arg):
        return sg.not_(self.cast(arg, dt.boolean))

    def visit_Date(self, op, *, arg):
        return self.cast(arg, dt.date)

    def visit_TryCast(self, op, *, arg, to):
        type_mapper = self.type_mapper
        if op.arg.dtype.is_temporal() and to.is_numeric():
            return self.f.unix_timestamp(
                sge.TryCast(this=arg, to=type_mapper.from_ibis(dt.string))
            )
        return sge.TryCast(this=arg, to=type_mapper.from_ibis(to))

    def visit_FloorDivide(self, op, *, left, right):
        return self.f.floor(left / right)

    def visit_JSONGetItem(self, op, *, arg, index):
        assert isinstance(op.index, ops.Literal)
        idx = op.index
        val = idx.value
        if idx.dtype.is_integer():
            query_path = f"$[{val}]"
        else:
            assert idx.dtype.is_string(), idx.dtype
            query_path = f"$.{val}"

        key_hack = f"{sge.convert(query_path).sql(self.dialect)} WITH CONDITIONAL ARRAY WRAPPER"
        return self.f.json_query(arg, self.v[key_hack])

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        from xorq.vendor.ibis.common.temporal import TimestampUnit

        if unit == TimestampUnit.MILLISECOND:
            precision = 3
        elif unit == TimestampUnit.SECOND:
            precision = 0
        else:
            raise ValueError(f"{unit!r} unit is not supported!")

        return self.cast(self.f.to_timestamp_ltz(arg, precision), dt.timestamp)

    def visit_Time(self, op, *, arg):
        return self.cast(arg, op.dtype)

    def visit_TimeFromHMS(self, op, *, hours, minutes, seconds):
        padded_hour = self.f.lpad(self.cast(hours, dt.string), 2, "0")
        padded_minute = self.f.lpad(self.cast(minutes, dt.string), 2, "0")
        padded_second = self.f.lpad(self.cast(seconds, dt.string), 2, "0")
        return self.cast(
            self.f.concat(padded_hour, ":", padded_minute, ":", padded_second), op.dtype
        )

    def visit_DateFromYMD(self, op, *, year, month, day):
        padded_year = self.f.lpad(self.cast(year, dt.string), 4, "0")
        padded_month = self.f.lpad(self.cast(month, dt.string), 2, "0")
        padded_day = self.f.lpad(self.cast(day, dt.string), 2, "0")
        return self.cast(
            self.f.concat(padded_year, "-", padded_month, "-", padded_day), op.dtype
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        padded_year = self.f.lpad(self.cast(year, dt.string), 4, "0")
        padded_month = self.f.lpad(self.cast(month, dt.string), 2, "0")
        padded_day = self.f.lpad(self.cast(day, dt.string), 2, "0")
        padded_hour = self.f.lpad(self.cast(hours, dt.string), 2, "0")
        padded_minute = self.f.lpad(self.cast(minutes, dt.string), 2, "0")
        padded_second = self.f.lpad(self.cast(seconds, dt.string), 2, "0")
        return self.cast(
            self.f.concat(
                padded_year,
                "-",
                padded_month,
                "-",
                padded_day,
                " ",
                padded_hour,
                ":",
                padded_minute,
                ":",
                padded_second,
            ),
            op.dtype,
        )

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.unix_timestamp(self.cast(arg, dt.string))

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype
        if to.is_timestamp():
            if from_.is_numeric():
                arg = self.f.from_unixtime(arg)
            if (tz := to.timezone) is not None:
                return self.f.to_timestamp(
                    self.f.convert_tz(self.cast(arg, dt.string), "UTC+0", tz)
                )
            else:
                return self.f.to_timestamp(arg, "yyyy-MM-dd HH:mm:ss.SSS")
        elif to.is_json():
            return arg
        elif from_.is_temporal() and to.is_int64():
            return 1_000_000 * self.f.unix_timestamp(arg)
        else:
            return self.cast(arg, to)

    def visit_IfElse(self, op, *, bool_expr, true_expr, false_null_expr):
        return self.if_(
            bool_expr,
            true_expr if true_expr != NULL else self.cast(true_expr, op.dtype),
            (
                false_null_expr
                if false_null_expr != NULL
                else self.cast(false_null_expr, op.dtype)
            ),
        )

    def visit_Log10(self, op, *, arg):
        return self.f.anon.log(10, arg)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.f.extract(self.v.millisecond, arg)

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.extract(self.v.microsecond, arg)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return (self.f.dayofweek(arg) + 5) % 7

    def visit_DayOfWeekName(self, op, *, arg):
        index = self.cast(self.f.dayofweek(self.cast(arg, dt.date)), op.dtype)
        lookup_table = self.f.str_to_map(
            "1=Sunday,2=Monday,3=Tuesday,4=Wednesday,5=Thursday,6=Friday,7=Saturday"
        )
        return lookup_table[index]

    def visit_TimestampNow(self, op):
        return self.v.current_timestamp

    def visit_DateNow(self, op):
        return self.v.current_date

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        unit = op.interval.dtype.unit.name
        unit_var = self.v[unit]

        if offset is None:
            offset = 0
        else:
            offset = op.offset.value

        bucket_width = op.interval.value
        unit_func = self.f["dayofmonth" if unit.upper() == "DAY" else unit]

        arg = self.f.anon.timestampadd(unit_var, -sge.paren(offset, copy=False), arg)
        mod = unit_func(arg) % bucket_width

        return self.f.anon.timestampadd(
            unit_var,
            -sge.paren(mod, copy=False) + offset,
            self.v[f"FLOOR({arg.sql(self.dialect)} TO {unit_var.sql(self.dialect)})"],
        )

    def visit_TemporalDelta(self, op, *, part, left, right):
        right = self.visit_TemporalTruncate(None, arg=right, unit=part)
        left = self.visit_TemporalTruncate(None, arg=left, unit=part)
        return self.f.anon.timestampdiff(
            self.v[part.this],
            self.cast(right, dt.timestamp),
            self.cast(left, dt.timestamp),
        )

    visit_TimeDelta = visit_DateDelta = visit_TimestampDelta = visit_TemporalDelta

    def visit_TemporalTruncate(self, op, *, arg, unit):
        unit_var = self.v[unit.name]
        arg_sql = arg.sql(self.dialect)
        unit_sql = unit_var.sql(self.dialect)
        return self.f.floor(self.v[f"{arg_sql} TO {unit_sql}"])

    visit_TimestampTruncate = visit_DateTruncate = visit_TimeTruncate = (
        visit_TemporalTruncate
    )

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.instr(haystack, needle) > 0

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise com.UnsupportedOperationError(
                "String find doesn't support `end` argument"
            )

        if start is not None:
            arg = self.f.substr(arg, start + 1)
            pos = self.f.instr(arg, substr)
            return self.if_(pos > 0, pos + start, 0)

        return self.f.instr(arg, substr)

    def visit_StartsWith(self, op, *, arg, start):
        return self.f.left(arg, self.f.char_length(start)).eq(start)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.char_length(end)).eq(end)

    def visit_ExtractUrlField(self, op, *, arg):
        return self.f.parse_url(arg, type(op).__name__[len("Extract") :].upper())

    visit_ExtractAuthority = visit_ExtractHost = visit_ExtractUserInfo = (
        visit_ExtractProtocol
    ) = visit_ExtractFile = visit_ExtractPath = visit_ExtractUrlField

    def visit_ExtractQuery(self, op, *, arg, key):
        return self.f.parse_url(*filter(None, (arg, "QUERY", key)))

    def visit_ExtractFragment(self, op, *, arg):
        return self.f.parse_url(arg, "REF")

    def visit_CountStar(self, op, *, arg, where):
        if where is None:
            return self.f.count(STAR)
        return self.f.sum(self.cast(where, dt.int64))

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg, self.f.array(arg)[2])
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_MapContains(self, op: ops.MapContains, *, arg, key):
        key_type = op.arg.dtype.key_type
        return self.f.array_contains(
            self.cast(self.f.map_keys(arg), dt.Array(value_type=key_type)),
            self.cast(key, key_type),
        )

    def visit_Map(self, op: ops.Map, *, keys, values):
        return self.cast(self.f.map_from_arrays(keys, values), op.dtype)

    def visit_MapMerge(self, op: ops.MapMerge, *, left, right):
        left_keys = self.f.map_keys(left)
        left_values = self.f.map_values(left)

        right_keys = self.f.map_keys(right)
        right_values = self.f.map_values(right)

        keys = self.f.array_concat(left_keys, right_keys)
        values = self.f.array_concat(left_values, right_values)

        return self.cast(self.f.map_from_arrays(keys, values), op.dtype)

    def visit_StructColumn(self, op, *, names, values):
        return self.cast(sge.Struct(expressions=list(values)), op.dtype)

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )
        # the only way to get filtering *and* respecting nulls is to use
        # `FILTER` syntax, but it's broken in various ways for other aggregates
        out = self.f.array_agg(arg)
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        if where is not None:
            out = sge.Filter(this=out, expression=sge.Where(this=where))
        return out

    def visit_Strip(self, op, *, arg):
        # TODO: at some point, the upstream `BTRIM` function should work, but it
        # currently doesn't, so we use a combination of left and right trim here
        return self.visit_RStrip(op, arg=self.visit_LStrip(op, arg=arg))


compiler = FlinkCompiler()



import toolz


@toolz.curry
def schema_contains(required, schema_in):
    return not set(required.items()).difference(schema_in.items())


@toolz.curry
def schema_concat(schema_in, to_concat):
    return schema_in | to_concat



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any

from public import public

import xorq.vendor.ibis.expr.operations as ops
from xorq.common.exceptions import (
    ExpressionError,
    InputTypeError,
    IntegrityError,
    XorqInputError,
)
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.deferred import Deferred
from xorq.vendor.ibis.common.egraph import DisjointSet
from xorq.vendor.ibis.expr.rewrites import flatten_predicates, peel_join_field
from xorq.vendor.ibis.expr.types.generic import Value
from xorq.vendor.ibis.expr.types.relations import (
    DerefMap,
    Table,
    bind,
    unwrap_aliases,
)


if TYPE_CHECKING:
    from collections.abc import Sequence

    from xorq.vendor.ibis.expr.operations.relations import JoinKind


def coerce_to_table(data):
    try:
        import pandas as pd
    except ImportError:
        pass
    else:
        if isinstance(data, pd.DataFrame):
            return ibis.memtable(data)

    try:
        import pyarrow as pa
    except ImportError:
        pass
    else:
        if isinstance(data, pa.Table):
            return ibis.memtable(data)

    if not isinstance(data, Table):
        raise TypeError(f"right operand must be a Table, got {type(data).__name__}")
    return data


def disambiguate_fields(
    how,
    predicates,
    equalities,
    left_fields,
    right_fields,
    left_template,
    right_template,
):
    """Resolve name collisions between the left and right tables."""
    collisions = set()
    left_template = left_template or "{name}"
    right_template = right_template or "{name}"

    if how == "inner" and util.all_of(predicates, ops.Equals):
        # for inner joins composed exclusively of equality predicates, we can
        # avoid renaming columns with colliding names if their values are
        # guaranteed to be equal due to the predicate
        equalities = equalities.copy()
        for pred in predicates:
            if isinstance(pred.left, ops.Field) and isinstance(pred.right, ops.Field):
                # disjoint sets are used to track the equality groups
                equalities.add(pred.left)
                equalities.add(pred.right)
                equalities.union(pred.left, pred.right)

    if how in ("semi", "anti"):
        # discard the right fields per left semi and left anty join semantics
        return left_fields, collisions, equalities

    fields = {}
    for name, field in left_fields.items():
        if name in right_fields:
            # there is an overlap between this field and a field from the right
            try:
                # check if the fields are equal due to equality predicates
                are_equal = equalities.connected(field, right_fields[name])
            except KeyError:
                are_equal = False
            if not are_equal:
                # there is a name collision and the fields are not equal, so
                # rename the field from the left according to the provided
                # template (which is the name itself by default)
                name = left_template.format(name=name)

        fields[name] = field

    for name, field in right_fields.items():
        if name in left_fields:
            # there is an overlap between this field and a field from the left
            try:
                # check if the fields are equal due to equality predicates
                are_equal = equalities.connected(field, left_fields[name])
            except KeyError:
                are_equal = False

            if are_equal:
                # even though there is a name collision, the fields are equal
                # due to equality predicates, so we can safely discard the
                # field from the right
                continue
            else:
                # there is a name collision and the fields are not equal, so
                # rename the field from the right according to the provided
                # template
                name = right_template.format(name=name)

        if name in fields:
            # we can still have collisions after multiple joins, or a wrongly
            # chosen template, so we need to track the collisions
            collisions.add(name)
        else:
            # the field name does not collide with any field from the left
            # and not occupied by any field from the right, so add it to the
            # fields mapping
            fields[name] = field

    return fields, collisions, equalities


def prepare_predicates(
    chain: ops.JoinChain,
    right: ops.Relation,
    predicates: Sequence[Any],
    comparison: type[ops.Comparison] = ops.Equals,
):
    """Bind and dereference predicates to the left and right tables.

    The responsibility of this function is twofold:
    1. Convert the various input values to valid predicates, including binding.
    2. Dereference the predicates one of the ops.JoinTable(s) in the join chain
       or the new JoinTable wrapping the right table. JoinTable(s) are used to
       ensure that all join participants are unique, even if the same table is
       joined multiple times.

    Since join predicates can be ambiguous sometimes, we do the two steps above
    in the same time so that we have more contextual information to resolve
    ambiguities.

    Possible inputs for the predicates:
    1. A python boolean literal, which is converted to a literal expression
    2. A boolean `Value` expression, which gets flattened and dereferenced.
       If there are comparison expressions where both sides depend on the same
       relation, then the left side is dereferenced to one of the join tables
       already part of the join chain, while the right side is dereferenced to
       the new join table wrapping the right table.
    3. A `Deferred` expression, which gets resolved on the left table and then
       the same path is followed as for `Value` expressions.
    4. A pair of expression-like objects, which are getting bound to the left
       and right tables respectively using the robust `bind` function handling
       several cases, including `Deferred` expressions, `Selector`s, literals,
       etc. Then the left are dereferenced to the join chain whereas the right
       to the new join table wrapping the right table.

    Parameters
    ----------
    chain
        The join chain
    right
        The right table
    predicates
        Predicates to bind and dereference, see the possible values above
    comparison
        The comparison operation to construct if the input is a pair of
        expression-like objects
    """
    reverse = {
        ops.Field(chain, k): v
        for k, v in chain.values.items()
        if isinstance(v, ops.Field)
    }
    deref_right = DerefMap.from_targets(right)
    deref_left = DerefMap.from_targets(chain.tables, extra=reverse)
    deref_both = DerefMap.from_targets([*chain.tables, right], extra=reverse)

    left, right = chain.to_expr(), right.to_expr()
    for pred in util.promote_list(predicates):
        if isinstance(pred, (Value, Deferred, bool)):
            for bound in bind(left, pred):
                yield deref_both.dereference(bound.op())
        else:
            if isinstance(pred, tuple):
                if len(pred) != 2:
                    raise ExpressionError("Join key tuple must be length 2")
                lk, rk = pred
            else:
                lk = rk = pred

            for lhs, rhs in zip(bind(left, lk), bind(right, rk)):
                lhs = deref_left.dereference(lhs.op())
                rhs = deref_right.dereference(rhs.op())
                yield comparison(lhs, rhs)


def finished(method):
    """Decorator to ensure the join chain is finished before calling a method."""

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        return method(self._finish(), *args, **kwargs)

    return wrapper


@public
class Join(Table):
    __slots__ = ("_collisions", "_equalities")

    def __init__(self, arg, collisions=(), equalities=()):
        assert isinstance(arg, ops.Node)
        if not isinstance(arg, ops.JoinChain):
            # coerce the input node to a join chain operation
            arg = ops.JoinReference(arg, identifier=0)
            arg = ops.JoinChain(arg, rest=(), values=arg.fields)
        super().__init__(arg)
        # the collisions and equalities are used to track the name collisions
        # and the equality groups join fields based on equality predicates;
        # these must be tracked in the join expression because the join chain
        # operation doesn't hold any information about `lname` and `rname`
        # parameters passed to the join methods and used to disambiguate field
        # names; the collisions are used to raise an error if there are any
        # name collisions after the join chain is finished
        object.__setattr__(self, "_collisions", collisions or set())
        object.__setattr__(self, "_equalities", equalities or DisjointSet())

    def _finish(self) -> Table:
        """Construct a valid table expression from this join expression."""
        if self._collisions:
            raise IntegrityError(f"Name collisions: {self._collisions}")
        return Table(self.op())

    @functools.wraps(Table.join)
    def join(
        self,
        right,
        predicates: Any,
        how: JoinKind = "inner",
        *,
        lname: str = "",
        rname: str = "{name}_right",
    ):
        right = coerce_to_table(right)

        if how == "left_semi":
            how = "semi"
        elif how == "asof":
            raise XorqInputError("use table.asof_join(...) instead")

        chain = self.op()
        right = right.op()
        if not isinstance(right, ops.Reference):
            right = ops.JoinReference(right, identifier=chain.length)

        # bind and dereference the predicates
        preds = prepare_predicates(chain, right, predicates)
        preds = flatten_predicates(preds)
        if not preds and how not in {"cross", "positional"}:
            # if there are no predicates, default to every row matching unless
            # the join is a cross join, because a cross join already has this
            # behavior
            preds.append(ops.Literal(True, dtype="bool"))

        # calculate the fields based in lname and rname, this should be a best
        # effort to avoid collisions, but does not raise if there are any
        # if no disambiaution happens using a final .select() call, then
        # the finish() method will raise due to the name collisions
        values, collisions, equalities = disambiguate_fields(
            how=how,
            predicates=preds,
            equalities=self._equalities,
            left_fields=chain.values,
            right_fields=right.fields,
            left_template=lname,
            right_template=rname,
        )

        # construct a new join link and add it to the join chain
        link = ops.JoinLink(how, table=right, predicates=preds)
        chain = chain.copy(rest=chain.rest + (link,), values=values)

        # return with a new JoinExpr wrapping the new join chain
        return self.__class__(chain, collisions=collisions, equalities=equalities)

    @functools.wraps(Table.asof_join)
    def asof_join(
        self: Table,
        right: Table,
        on,
        predicates=(),
        tolerance=None,
        *,
        lname: str = "",
        rname: str = "{name}_right",
    ):
        predicates = util.promote_list(predicates)
        if tolerance is not None:
            # `tolerance` parameter is mimicking the pandas API, but we express
            # it at the expression level by a sequence of operations:
            # 1. perform the `asof` join with the `on` an `predicates` parameters
            #    where the `on` parameter is an inequality predicate
            # 2. filter the asof join result using the `tolerance` parameter and
            #    the `on` parameter
            # 3. perform a left join between the original left table and the
            #    filtered asof join result using the left operand of the `on` parameter but this
            #    time as an equality predicate
            if isinstance(on, str):
                # self is always a JoinChain so reference one of the join tables
                left_on = self.op().values[on].to_expr()
                right_on = right[on]
                on = left_on >= right_on
            elif isinstance(on, Value):
                node = on.op()
                if not isinstance(node, ops.Binary):
                    raise InputTypeError("`on` must be a comparison expression")
                left_on = node.left.to_expr()
                right_on = node.right.to_expr()
            else:
                raise TypeError("`on` must be a string or a ValueExpr")

            joined = self.asof_join(
                right, on=on, predicates=predicates, lname=lname, rname=rname
            )
            filtered = joined.filter(
                left_on <= right_on + tolerance, left_on >= right_on - tolerance
            )
            (right_on,) = filtered.bind(left_on)

            # without joining twice the table would not contain the rows from
            # the left table that do not match any row from the right table
            # given the tolerance filter
            result = self.left_join(
                filtered, predicates=[left_on == right_on] + predicates
            )
            values = {**filtered.op().values, **self.op().values}

            return result.select(**values)

        chain = self.op()
        right = right.op()
        if not isinstance(right, ops.Reference):
            right = ops.JoinReference(right, identifier=chain.length)

        # TODO(kszucs): add extra validation for `on` with clear error messages
        (on,) = prepare_predicates(chain, right, [on], comparison=ops.GreaterEqual)
        preds = prepare_predicates(chain, right, predicates, comparison=ops.Equals)
        preds = [on, *preds]

        values, collisions, equalities = disambiguate_fields(
            how="asof",
            predicates=preds,
            equalities=self._equalities,
            left_fields=chain.values,
            right_fields=right.fields,
            left_template=lname,
            right_template=rname,
        )

        # construct a new join link and add it to the join chain
        link = ops.JoinLink("asof", table=right, predicates=preds)
        chain = chain.copy(rest=chain.rest + (link,), values=values)

        # return with a new JoinExpr wrapping the new join chain
        return self.__class__(chain, collisions=collisions, equalities=equalities)

    @functools.wraps(Table.cross_join)
    def cross_join(
        self: Table,
        right: Table,
        *rest: Table,
        lname: str = "",
        rname: str = "{name}_right",
    ):
        left = self.join(right, how="cross", predicates=(), lname=lname, rname=rname)
        for table in rest:
            left = left.join(
                table, how="cross", predicates=(), lname=lname, rname=rname
            )
        return left

    @functools.wraps(Table.select)
    def select(self, *args, **kwargs):
        chain = self.op()
        values = self.bind(*args, **kwargs)
        values = unwrap_aliases(values)

        links = [link.table for link in chain.rest if link.how not in ("semi", "anti")]
        derefmap = DerefMap.from_targets([chain.first, *links])

        # if there are values referencing fields from the join chain constructed
        # so far, we need to replace them the fields from one of the join links
        values = {
            k: v.replace(peel_join_field, filter=ops.Value) for k, v in values.items()
        }
        values = {k: derefmap.dereference(v) for k, v in values.items()}

        node = chain.copy(values=values)
        return Table(node)

    aggregate = finished(Table.aggregate)
    alias = finished(Table.alias)
    cast = finished(Table.cast)
    compile = finished(Table.compile)
    count = finished(Table.count)
    difference = finished(Table.difference)
    distinct = finished(Table.distinct)
    drop = finished(Table.drop)
    dropna = finished(Table.dropna)
    execute = finished(Table.execute)
    fill_null = finished(Table.fill_null)
    filter = finished(Table.filter)
    group_by = finished(Table.group_by)
    intersect = finished(Table.intersect)
    limit = finished(Table.limit)
    mutate = finished(Table.mutate)
    nunique = finished(Table.nunique)
    order_by = finished(Table.order_by)
    sample = finished(Table.sample)
    sql = finished(Table.sql)
    unbind = finished(Table.unbind)
    union = finished(Table.union)
    view = finished(Table.view)


public(JoinExpr=Join)



import tarfile
import tempfile
from contextlib import contextmanager
from pathlib import Path

from xorq.catalog.constants import (
    PREFERRED_SUFFIX,
)
from xorq.ibis_yaml.compiler import DumpFiles


REQUIRED_TGZ_NAMES = (DumpFiles.expr, DumpFiles.metadata, DumpFiles.profiles)


def test_tgz(tgz_path):
    with tarfile.TarFile.gzopen(tgz_path) as tfh:
        # https://docs.python.org/3/library/tarfile.html#cmdoption-tarfile-t
        relpaths = tuple(
            Path(member.path) for member in tfh.getmembers() if member.isfile()
        )
        missing = set(REQUIRED_TGZ_NAMES).difference(path.name for path in relpaths)
        assert not missing, missing


@contextmanager
def make_tgz_context(build_dir):
    path_prefix = (build_dir := Path(build_dir)).parent
    with tempfile.TemporaryDirectory() as td:
        tgz_path = Path(td).joinpath(build_dir.name).with_suffix(PREFERRED_SUFFIX)
        with tarfile.TarFile.gzopen(tgz_path, "w") as tf:
            for p in build_dir.iterdir():
                tf.add(p, arcname=p.relative_to(path_prefix))
        yield tgz_path


@contextmanager
def extract_build_tgz_context(tgz_path):
    with tempfile.TemporaryDirectory() as td:
        with tarfile.TarFile.gzopen(tgz_path) as tfh:
            tfh.extractall(td, filter="data")
        (path, *rest) = Path(td).iterdir()
        assert not rest
        yield path


def write_tgz(path, relpath_to_bytes):
    with tempfile.TemporaryDirectory() as td:
        with tarfile.TarFile.gzopen((path := Path(path)), "w") as tfh:
            for relpath, byts in dict(relpath_to_bytes).items():
                p = Path(td).joinpath(relpath)
                p.parent.mkdir(exist_ok=True, parents=True)
                p.write_bytes(byts)
                tfh.add(p, arcname=relpath)
    return path



import operator
import pickle

import pyarrow as pa
import pytest
import toolz

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.api import make_pandas_expr_udf, make_pandas_udf, udf
from xorq.expr.ml.enums import ResponseMethod
from xorq.tests.util import assert_frame_equal
from xorq.vendor.ibis import _
from xorq.vendor.ibis import literal as L
from xorq.vendor.ibis.selectors import of_type


pc = pytest.importorskip("pyarrow.compute")
xgb = pytest.importorskip("xgboost")


@udf.agg.pyarrow
def my_mean(arr: dt.float64) -> dt.float64:
    return pc.mean(arr)


@udf.agg.pyarrow
def add_mean(a: dt.float64, b: dt.float64) -> dt.float64:
    return pc.mean(pc.add(a, b))


return_type = pa.struct(
    [
        ("a", pa.float64()),
        ("b", pa.float64()),
        ("c", pa.float64()),
    ]
)


@udf.agg.pyarrow
def centroid(a: dt.float64, b: dt.float64, c: dt.float64) -> return_type:
    return pa.scalar((pc.mean(a), pc.mean(b), pc.mean(c)), type=return_type)


@udf.agg.pyarrow
def centroid_list(a: dt.float64, b: dt.float64, c: dt.float64) -> pa.list_(
    pa.float64()
):
    return pa.scalar([pc.mean(a), pc.mean(b), pc.mean(c)], type=pa.list_(pa.float64()))


def test_udf_agg_pyarrow(ls_con, batting):
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    result = my_mean(batting.G).execute()

    assert result == xo.execute(batting.G).mean()


def test_multiple_arguments_udf_agg_pyarrow(ls_con, batting):
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = add_mean(batting.G, batting.G).execute()
    expected = xo.execute(batting.G)
    expected = (expected + expected).mean()

    assert actual == expected


def test_multiple_arguments_struct_udf_agg_pyarrow(ls_con, batting):
    from math import isclose

    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = centroid(batting.G, batting.G, batting.G).execute()
    expected = xo.execute(batting.G).mean()

    assert all(isclose(value, expected) for value in actual.values())


def test_multiple_arguments_list_udf_agg_pyarrow(ls_con, batting):
    from math import isclose

    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = centroid_list(batting.G, batting.G, batting.G).execute()
    expected = xo.execute(batting.G).mean()

    assert all(isclose(value, expected) for value in actual)


@udf.agg.pyarrow
def my_sum(arr: dt.float64) -> dt.float64:
    return pc.sum(arr)


def test_group_by_udf_agg_pyarrow(ls_con, alltypes_df):
    alltypes = ls_con.register(alltypes_df, "pg-alltypes")

    expr = (
        alltypes[_.string_col == "1"]
        .mutate(x=L(1, "int64"))
        .group_by(_.x)
        .aggregate(sum=my_sum(_.double_col))
    )

    result = xo.execute(expr).astype({"x": "int64"})
    expected = (
        alltypes_df.loc[alltypes_df.string_col == "1", :]
        .assign(x=1)
        .groupby("x")
        .double_col.sum()
        .rename("sum")
        .reset_index()
    )

    assert_frame_equal(result, expected, check_like=True)


def test_udf_agg_pandas_df(ls_con, alltypes):
    name = "sum_sum"
    alltypes = ls_con.register(xo.execute(alltypes), "pg-alltypes")
    cols = (by, _) = ["year", "month"]
    expr = alltypes

    @udf.agg.pandas_df(
        schema=expr[cols].schema(),
        return_type=dt.int64(),
        name=name,
    )
    def sum_sum(df):
        return df.sum().sum()

    actual = (
        expr.group_by(by)
        .agg(sum_sum(*(expr[c] for c in cols)).name(name))
        .execute()
        .sort_values(by, ignore_index=True)
    )
    actual2 = (
        expr.group_by(by)
        .agg(sum_sum.on_expr(expr).name(name))
        .execute()
        .sort_values(by, ignore_index=True)
    )
    expected = (
        alltypes[cols]
        .execute()
        .groupby(by)
        .apply(sum_sum.fn)
        .rename(name)
        .reset_index()
        .sort_values(by, ignore_index=True)
    )
    assert actual.equals(expected)
    assert actual2.equals(expected)


def test_udf_agg_pandas_df_no_name():
    import xorq.api as xo

    @xo.udf.agg.pandas_df(
        schema=xo.schema({"a": int}),
        return_type=xo.expr.datatypes.Int64(),
    )
    def my_agg(df):
        return df.a.sum()

    @xo.udf.agg.pandas_df(
        schema=xo.schema({"a": int}),
        return_type=xo.expr.datatypes.Int64(),
        name="my_agg",  # the name clashes with the previous one
    )
    def my_agg_plus_one(df):
        return df.a.sum() + 1

    t = xo.memtable({"a": [1, 2]})
    expr = t.pipe(my_agg.on_expr)
    assert expr.execute() == 3

    expr = t.pipe(my_agg_plus_one.on_expr)
    assert expr.execute() == 4

    expr = t.mutate(my_agg=my_agg(t.a), my_agg_plus_one=my_agg_plus_one(t.a))
    assert not expr.execute().empty


def test_udf_pandas_df(ls_con, batting):
    typ = "int64"
    name = "my_least"
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    my_least = operator.methodcaller("min", axis=1)
    schema = batting.select(of_type(typ)).schema()
    return_type = dt.dtype(typ)
    udf = make_pandas_udf(my_least, schema, return_type, name=name)
    order_by = ("playerID", "yearID", "stint")

    from_builtin = ls_con.execute(
        batting.mutate(xo.least(*(xo._[name] for name in schema)).name(name)).order_by(
            order_by
        )
    )
    from_udf = ls_con.execute(
        batting.mutate(udf.on_expr(batting).name(name)).order_by(order_by)
    )
    from_pd = (
        ls_con.execute(batting)
        .assign(**{name: lambda t: my_least(t[list(schema)])})
        .sort_values(list(order_by), ignore_index=True)
    )
    assert from_builtin.equals(from_udf)
    assert from_builtin.equals(from_pd)


def test_pandas_expr_udf(con, diamonds):
    @toolz.curry
    def train_xgboost_model(df, features, target, seed=0):
        param = {
            "max_depth": 3,
            "eta": 1,
            "objective": "reg:squarederror",
            "seed": seed,
        }
        num_round = 5
        X = df[list(features)]
        y = df[target]
        dtrain = xgb.DMatrix(X, y)
        bst = xgb.train(param, dtrain, num_boost_round=num_round)
        return bst

    def predict_xgboost_model(model, df):
        return model.predict(xgb.DMatrix(df))

    features = (
        "carat",
        "depth",
        "table",
        "x",
        "y",
        "z",
    )
    target = "price"
    train_fn = train_xgboost_model(features=features, target=target)
    name = ResponseMethod.PREDICT
    typ = "float64"

    df = diamonds.limit(500).execute()
    t = con.register(df, table_name="small_diamonds")

    # manual run
    model = train_fn(df)
    from_pd = df.assign(
        **{name: predict_xgboost_model(model, df[list(features)])}
    ).astype({name: typ})

    # using expr-scalar-udf
    model_udaf = udf.agg.pandas_df(
        fn=toolz.compose(pickle.dumps, train_fn),
        schema=t[features + (target,)].schema(),
        return_type=dt.binary,
        name="xgboost_model",
    )
    computed_kwargs_expr = model_udaf.on_expr(t)
    predict_expr_udf = make_pandas_expr_udf(
        computed_kwargs_expr,
        fn=predict_xgboost_model,
        schema=t[features].schema(),
        return_type=dt.dtype(typ),
        name=name,
    )

    expr = t.mutate(predict_expr_udf.on_expr(t).name(name))
    from_ls = con.execute(expr)

    assert_frame_equal(from_ls, from_pd)



from xorq.config import _backend_init
from xorq.examples.core import (
    get_name_to_suffix,
    get_table_from_name,
    whitelist,
)


class Example:
    def __init__(self, name):
        self.name = name

    def fetch(self, backend=None, table_name=None, deferred=True, **kwargs):
        if backend is None:
            backend = _backend_init()

        return get_table_from_name(
            self.name,
            backend,
            table_name or self.name,
            deferred=deferred,
            **kwargs,
        )


def __dir__():
    return (
        "get_table_from_name",
        *whitelist,
    )


def __getattr__(name):
    from xorq.vendor.ibis import examples as ibex

    lookup = get_name_to_suffix()

    if name not in lookup:
        return getattr(ibex, name)

    return Example(name)


__all__ = (
    "get_table_from_name",
    *whitelist,
)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any

import pandas as pd
import pyarrow as pa
import pyarrow_hotfix  # noqa: F401

import xorq.common.exceptions as com
import xorq.vendor.ibis.config
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends import BaseBackend, NoUrl
from xorq.vendor.ibis.common.dispatch import lazy_singledispatch
from xorq.vendor.ibis.formats.pandas import PandasData, PandasSchema
from xorq.vendor.ibis.formats.pyarrow import PyArrowData


if TYPE_CHECKING:
    import pathlib
    from collections.abc import Mapping, MutableMapping


class BasePandasBackend(BaseBackend, NoUrl):
    """Base class for backends based on pandas."""

    name = "pandas"
    dialect = None
    backend_table_type = pd.DataFrame

    class Options(xorq.vendor.ibis.config.Config):
        enable_trace: bool = False

    def do_connect(
        self,
        dictionary: MutableMapping[str, pd.DataFrame] | None = None,
    ) -> None:
        """Construct a client from a dictionary of pandas DataFrames.

        Parameters
        ----------
        dictionary
            An optional mapping of string table names to pandas DataFrames.

        Examples
        --------
        >>> import xorq.api as xo
        >>> xo.pandas.connect({"t": pd.DataFrame({"a": [1, 2, 3]})})  # doctest: +ELLIPSIS
        <ibis.backends.pandas.Backend object at 0x...>
        """
        self.dictionary = dictionary or {}
        self.schemas = {}

    def disconnect(self) -> None:
        pass

    def from_dataframe(
        self,
        df: pd.DataFrame,
        name: str = "df",
        client: BasePandasBackend | None = None,
    ) -> ir.Table:
        """Construct an ibis table from a pandas DataFrame.

        Parameters
        ----------
        df
            A pandas DataFrame
        name
            The name of the pandas DataFrame
        client
            Client dictionary will be mutated with the name of the DataFrame,
            if not provided a new client is created

        Returns
        -------
        Table
            A table expression

        """
        if client is None:
            return self.connect({name: df}).table(name)
        client.dictionary[name] = df
        return client.table(name)

    def read_csv(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a CSV file as a table in the current session.

        Parameters
        ----------
        source
            The data source. Can be a local or remote file, pathlike objects
            also accepted.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_csv")
        df = pd.read_csv(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    def read_parquet(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a parquet file as a table in the current session.

        Parameters
        ----------
        source
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_parquet")
        df = pd.read_parquet(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    @property
    def version(self) -> str:
        return pd.__version__

    def list_tables(self, like=None, database=None):
        """Return the list of table names in the current database.

        Parameters
        ----------
        like
            A pattern in Python's regex format.
        database
            Unused in the pandas backend.

        Returns
        -------
        list[str]
            The list of the table names that match the pattern `like`.
        """
        return self._filter_with_like(list(self.dictionary.keys()), like)

    def table(self, name: str, schema: sch.Schema | None = None):
        inferred_schema = self.get_schema(name)
        overridden_schema = {**inferred_schema, **(schema or {})}
        return ops.DatabaseTable(name, overridden_schema, self).to_expr()

    def get_schema(self, table_name, *, database=None):
        try:
            schema = self.schemas[table_name]
        except KeyError:
            df = self.dictionary[table_name]
            self.schemas[table_name] = schema = PandasData.infer_table(df)

        return schema

    def compile(self, expr, *args, **kwargs):
        return expr

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        """Create a table."""
        if temp:
            com.XorqError(
                "Passing `temp=True` to the Pandas backend create_table method has no "
                "effect: all tables are in memory and temporary."
            )
        if database:
            com.XorqError(
                "Passing `database` to the Pandas backend create_table method has no "
                "effect: Pandas cannot set a database."
            )
        if obj is None and schema is None:
            raise com.XorqError("The schema or obj parameter is required")
        if schema is not None:
            schema = ibis.schema(schema)

        if obj is not None:
            df = self._convert_object(obj)
        else:
            dtypes = dict(PandasSchema.from_ibis(schema))
            df = pd.DataFrame(columns=dtypes.keys()).astype(dtypes)

        if name in self.dictionary and not overwrite:
            raise com.XorqError(f"Cannot overwrite existing table `{name}`")

        self.dictionary[name] = df

        if schema is not None:
            self.schemas[name] = schema
        return self.table(name)

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        return self.create_table(
            name, obj=obj, temp=None, database=database, overwrite=overwrite
        )

    def drop_view(self, name: str, *, force: bool = False) -> None:
        self.drop_table(name, force=force)

    def drop_table(self, name: str, *, force: bool = False) -> None:
        try:
            del self.dictionary[name]
        except KeyError:
            if not force:
                raise com.XorqError(f"Table {name} does not exist") from None

    def _convert_object(self, obj: Any) -> Any:
        return _convert_object(obj, self)

    @classmethod
    @lru_cache
    def _get_operations(cls):
        from xorq.backends.pandas.kernels import supported_operations

        return supported_operations

    @classmethod
    def has_operation(cls, operation: type[ops.Value]) -> bool:
        return operation in cls._get_operations()

    def _drop_cached_table(self, name):
        del self.dictionary[name]

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        table_expr = expr.as_table()
        output = pa.Table.from_pandas(
            self.execute(table_expr, params=params, limit=limit, **kwargs)
        )

        # cudf.pandas adds a column with the name `__index_level_0__` (and maybe
        # other index level columns) but these aren't part of the known schema
        # so we drop them
        output = output.drop(
            filter(lambda col: col.startswith("__index_level_"), output.column_names)
        )
        table = PyArrowData.convert_table(output, table_expr.schema())
        return expr.__pyarrow_result__(table)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1000000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        pa = self._import_pyarrow()
        pa_table = self.to_pyarrow(
            expr.as_table(), params=params, limit=limit, **kwargs
        )
        return pa.ipc.RecordBatchReader.from_batches(
            pa_table.schema, pa_table.to_batches(max_chunksize=chunk_size)
        )


class Backend(BasePandasBackend):
    name = "pandas"

    def execute(self, query, params=None, limit="default", **kwargs):
        from xorq.backends.pandas.executor import PandasExecutor

        if limit != "default" and limit is not None:
            raise ValueError(
                "limit parameter to execute is not yet implemented in the "
                "pandas backend"
            )

        if not isinstance(query, ir.Expr):
            raise TypeError(
                f"`query` has type {type(query).__name__!r}, expected ibis.expr.types.Expr"
            )

        params = params or {}
        params = {k.op() if isinstance(k, ir.Expr) else k: v for k, v in params.items()}

        return PandasExecutor.execute(query.op(), backend=self, params=params)

    def _create_cached_table(self, name, expr):
        return self.create_table(name, expr.execute())

    def _finalize_memtable(self, name: str) -> None:
        """No-op, let Python handle clean up."""

    def drop_table(self, name: str, *, force: bool = False) -> None:
        if not force and name in self.dictionary:
            raise com.XorqError(
                "Cannot drop existing table. Call drop_table with force=True to drop existing table."
            )
        del self.dictionary[name]
        del self.schemas[name]

    def read_record_batches(
        self,
        record_batches: pa.RecordBatchReader
        | list[pa.RecordBatch]
        | tuple[pa.RecordBatch],
        table_name: str | None = None,
    ):
        if isinstance(record_batches, (list, tuple)):
            record_batches = pa.RecordBatchReader.from_batches(
                record_batches[0].schema, record_batches
            )

        self.dictionary[table_name] = record_batches.read_pandas()
        return self.table(table_name)


@lazy_singledispatch
def _convert_object(obj: Any, _conn):
    raise com.BackendConversionError(
        f"Unable to convert {obj.__class__} object "
        f"to backend type: {_conn.__class__.backend_table_type}"
    )


@_convert_object.register("ibis.expr.types.Table")
def _table(obj, _conn):
    if isinstance(op := obj.op(), ops.InMemoryTable):
        return op.data.to_frame()
    else:
        raise com.BackendConversionError(
            f"Unable to convert {obj.__class__} object "
            f"to backend type: {_conn.__class__.backend_table_type}"
        )


@_convert_object.register("polars.DataFrame")
@_convert_object.register("pyarrow.Table")
def _pa_polars(obj, _conn):
    return obj.to_pandas()


@_convert_object.register("polars.LazyFrame")
def _polars_lazy(obj, _conn):
    return obj.collect().to_pandas()


@_convert_object.register("pandas.DataFrame")
def _pandas(obj, _conn):
    return obj



"""Operations for working with maps."""

from __future__ import annotations

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.expr.operations.core import Unary, Value


@public
class Map(Value):
    """Construct a map."""

    keys: Value[dt.Array]
    values: Value[dt.Array]

    shape = rlz.shape_like("args")

    @attribute
    def dtype(self):
        return dt.Map(
            self.keys.dtype.value_type,
            self.values.dtype.value_type,
        )


@public
class MapLength(Unary):
    """Compute the number of unique keys in a map."""

    arg: Value[dt.Map]
    dtype = dt.int64


@public
class MapGet(Value):
    """Get a value from a map by key."""

    arg: Value[dt.Map]
    key: Value
    default: Value = None

    shape = rlz.shape_like("args")

    @attribute
    def dtype(self):
        return dt.higher_precedence(self.default.dtype, self.arg.dtype.value_type)


@public
class MapContains(Value):
    """Check if a map contains a key."""

    arg: Value[dt.Map]
    key: Value

    shape = rlz.shape_like("args")
    dtype = dt.bool


@public
class MapKeys(Unary):
    """Get the keys of a map as an array."""

    arg: Value[dt.Map]

    @attribute
    def dtype(self):
        return dt.Array(self.arg.dtype.key_type)


@public
class MapValues(Unary):
    """Get the values of a map as an array."""

    arg: Value[dt.Map]

    @attribute
    def dtype(self):
        return dt.Array(self.arg.dtype.value_type)


@public
class MapMerge(Value):
    """Combine two maps into one.

    If a key is present in both maps, the value from the first is kept.
    """

    left: Value[dt.Map]
    right: Value[dt.Map]

    shape = rlz.shape_like("args")
    dtype = rlz.dtype_like("args")


public(MapValueForKey=MapGet, MapValueOrDefaultForKey=MapGet, MapConcat=MapMerge)
